/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/
package org.jboss.remoting;

import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.jboss.logging.Logger;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.invocation.InternalInvocation;
import org.jboss.remoting.invocation.OnewayInvocation;
import org.jboss.remoting.marshal.Marshaller;
import org.jboss.remoting.marshal.UnMarshaller;
import org.jboss.remoting.stream.StreamServer;
import org.jboss.remoting.transport.ClientInvoker;
import org.jboss.util.id.GUID;
import org.jboss.util.threadpool.BasicThreadPool;
import org.jboss.util.threadpool.BlockingMode;
import org.jboss.util.threadpool.ThreadPool;

/**
 * Client is a convience method for invoking remote methods for a given subsystem
 *
 * @author <a href="mailto:jhaynie@vocalocity.net">Jeff Haynie</a>
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 * @version $Revision: 1.14 $
 */
public class Client
{
   public static final String LISTENER_ID_KEY = "listenerId";

   /**
    * Specifies the default number of work threads in the pool for
    * executing one way invocations on the client.
    * Value is 10.
    */
   public static final int MAX_NUM_ONEWAY_THREADS = 10;

   /**
    * Indicated the max number of threads used within oneway thread pool.
    */
   private int maxNumberThreads = MAX_NUM_ONEWAY_THREADS;
   private static final Logger log = Logger.getLogger(Client.class);
   private ClientInvoker invoker;
   private ClassLoader classloader;
   private String subsystem;
   private String sessionId = new GUID().toString();
   private ThreadPool onewayThreadPool;
   public static final String RAW = "RAW_PAYLOAD";

   private ConnectionValidator connectionValidator = null;

   public Client(InvokerLocator locator) throws Exception
   {
      this(locator, null);
   }

   public Client(InvokerLocator locator, String subsystem)
         throws Exception
   {
      this(Thread.currentThread().getContextClassLoader(), locator, subsystem);
   }

   public Client(ClassLoader cl, InvokerLocator locator, String subsystem)
         throws Exception
   {
      this(cl, InvokerRegistry.createClientInvoker(locator), subsystem);
   }

   public Client(ClassLoader cl, ClientInvoker invoker, String subsystem)
         throws Exception
   {
      this.classloader = cl;
      this.subsystem = subsystem == null ? null : subsystem.toUpperCase();
      this.invoker = invoker;
   }

   /**
    * Adds a connection listener that will be notified if/when the connection
    * to the server fails while the client is idle (no calls being made).
    *
    * @param listener
    */
   public void addConnectionListener(ConnectionListener listener)
   {
      if(connectionValidator == null)
      {
         connectionValidator = new ConnectionValidator(this);
      }
      connectionValidator.addConnectionListener(listener);
   }

   /**
    * Removes specified connection listener.  Will return true if it has
    * already been registered, false otherwise.
    *
    * @param listener
    * @return
    */
   public boolean removeConnectionListener(ConnectionListener listener)
   {
      return connectionValidator.removeConnectionListener(listener);
   }

   /**
    * This will set the session id used when making invocations on
    * server invokers.  There is a default unique id automatically
    * generated for each Client instance, so unless you have a good reason to set
    * this, do not set this.
    *
    * @param sessionId
    */
   public void setSessionId(String sessionId)
   {
      this.sessionId = sessionId;
   }

   public String getSessionId()
   {
      return this.sessionId;
   }

   public boolean isConnected()
   {
      return (this.invoker != null && this.invoker.isConnected());
   }

   public void connect() throws Exception
   {
      this.invoker.connect();
   }

   public void disconnect()
   {
      this.invoker.disconnect();
   }

   public ClientInvoker getInvoker()
   {
      return invoker;
   }

   public void setInvoker(ClientInvoker invoker)
   {
      this.invoker = invoker;
   }

   public String getSubsystem()
   {
      return subsystem;
   }

   public void setSubsystem(String subsystem)
   {
      this.subsystem = subsystem;
   }

   /**
    * Invokes the server invoker handler with the payload parameter passed.
    *
    * @param param
    * @return
    * @throws Throwable
    */
   public Object invoke(Object param) throws Throwable
   {
      return invoke(param, null);
   }

   /**
    * invoke the method remotely
    *
    * @param param    - payload for the server invoker handler
    * @param metadata - any extra metadata that may be needed by the transport (i.e. GET or POST if using
    *                 http invoker) or if need to pass along extra data to the server invoker handler.
    * @return
    * @throws Throwable
    */
   public Object invoke(Object param, Map metadata)
         throws Throwable
   {
      return invoke(param, metadata, null);
   }

   private Object invoke(Object param, Map metadata, InvokerLocator callbackServerLocator)
         throws Throwable
   {
      /**
       * Using a local variable for the invoker as work around so don't have
       * to sync method (and take performance hit)
       * Although this may cause having multiple instances of invoker in existance at
       * one time, should avoid having reference changed by another thread while in
       * execution path for method.
       */
      ClientInvoker localInvoker = invoker;

      if(localInvoker != null)
      {
         /**
          * This is here due to way the InvokerRegistry works.  Since it will cache references to
          * client invokers it creates based on locator, it is possible that this instance will
          * have a reference to the same instance of the invoker as another client.  So is possible
          * that client will disconnect, which will cause the invoker to be disconnected.  Therefore,
          * if this were to happen, we would create a new one by calling the createClientInvoker() method.
          */
         if(localInvoker.isConnected() == false)
         {
            log.debug("invoke called, but our invoker is disconnected, discarding and fetching another fresh invoker for: " + invoker.getLocator());

            localInvoker = InvokerRegistry.createClientInvoker(localInvoker.getLocator());
            localInvoker.connect();
         }
      }
      else
      {
         throw new Exception("Can not perform invoke because invoker is null.");
      }

      Object ret = localInvoker.invoke(new InvocationRequest(sessionId, subsystem, param, metadata, null, callbackServerLocator));
      this.invoker = localInvoker;
      return ret;
   }

   /**
    * Will invoke a oneway call to server without a return object.  This should be used when not expecting a
    * return value from the server and wish to achieve higher performance, since the client will not wait for
    * a return.
    * <b>
    * This is done one of two ways.  The first is to pass true as the clientSide param.  This will cause the
    * execution of the remote call to be excuted in a new thread on the client side and will return the calling thread
    * before making call to server side.  Although, this is optimal for performance, will not know about any problems
    * contacting server.
    * <p/>
    * The second, is to pass false as the clientSide param.  This will allow the current calling thread to make
    * the call to the remote server, at which point, the server side processing of the thread will be executed on
    * the remote server in a new executing thread and the client thread will return.  This is a little slower, but
    * will know that the call made it to the server.
    *
    * @param param
    * @param sendPayload
    * @param clientSide
    */
   public void invokeOneway(final Object param, final Map sendPayload, boolean clientSide) throws Throwable
   {
      if(clientSide)
      {
         ThreadPool threadPool = getOnewayThreadPool();
         Runnable onewayRun = new Runnable()
         {
            public void run()
            {
               try
               {
                  invoke(param, sendPayload);
               }
               catch(Throwable e)
               {
                  // throw away exception since can't get it back to original caller
                  log.error("Error executing client oneway invocation request: " + param, e);
               }
            }
         };
         threadPool.run(onewayRun);
      }
      else
      {
         OnewayInvocation invocation = new OnewayInvocation(param);
         invoke(invocation, sendPayload);
      }
   }

   public void setMaxNumberOfThreads(int numOfThreads)
   {
      this.maxNumberThreads = numOfThreads;
   }

   public int getMaxNumberOfThreads()
   {
      return this.maxNumberThreads;
   }

   public ThreadPool getOnewayThreadPool()
   {
      if(onewayThreadPool == null)
      {
         BasicThreadPool pool = new BasicThreadPool("JBossRemoting Client Oneway");
         pool.setMaximumPoolSize(maxNumberThreads);
         pool.setBlockingMode(BlockingMode.WAIT);
         onewayThreadPool = pool;
      }
      return onewayThreadPool;
   }

   public void setOnewayThreadPool(ThreadPool pool)
   {
      this.onewayThreadPool = pool;
   }


   /**
    * Same as calling invokeOneway(Object param, Map sendPayload, boolean clientSide) with
    * clientSide param being false.  Therefore, client thread will not return till it has made
    * remote call.
    *
    * @param param
    * @param sendPayload
    */
   public void invokeOneway(Object param, Map sendPayload) throws Throwable
   {
      invokeOneway(param, sendPayload, false);
   }

   /**
    * Adds the specified handler as a callback listener for pull (sync) callbacks.
    * The invoker server will then collect the callbacks for this specific handler.
    * The callbacks can be retrieved by calling the getCallbacks() method.
    * Note: this will cause the client invoker's client locator to be set to null.
    *
    * @param callbackHandler
    * @throws Throwable
    */
   public void addListener(InvokerCallbackHandler callbackHandler) throws Throwable
   {
      addListener(callbackHandler, null);
   }

   /**
    * Adds the specified handler as a callback listener for push (async) callbacks.
    * The invoker server will then callback on this handler (via the server invoker
    * specified by the clientLocator) when it gets a callback from the server handler.
    * Note: passing a null clientLocator will cause the client invoker's client
    * locator to be set to null.
    *
    * @param callbackHandler
    * @param clientLocator
    * @throws Throwable
    */
   public void addListener(InvokerCallbackHandler callbackHandler,
                           InvokerLocator clientLocator) throws Throwable
   {
      addListener(callbackHandler, clientLocator, null);
   }

   /**
    * Adds the specified handler as a callback listener for push (async) callbacks.
    * The invoker server will then callback on this handler (via the server invoker
    * specified by the clientLocator) when it gets a callback from the server handler.
    * Note: passing a null clientLocator will cause the client invoker's client
    * locator to be set to null.
    *
    * @param callbackHandler       interface to call on with callback
    * @param clientLocator         locator for callback server to callback on
    * @param callbackHandlerObject will be included in the callback object passed upon callback
    * @throws Throwable
    */
   public void addListener(InvokerCallbackHandler callbackHandler,
                           InvokerLocator clientLocator, Object callbackHandlerObject) throws Throwable
   {
      if(callbackHandler != null)
      {
         Map metadata = createListenerMetadata(callbackHandler);
         String listenerId = (String) metadata.get(LISTENER_ID_KEY);
         invoker.addClientLocator(listenerId, clientLocator);
         if(clientLocator != null)
         {
            Client client = new Client(clientLocator, subsystem);
            client.setSessionId(getSessionId());
            client.connect();

            client.invoke(new InternalInvocation(InternalInvocation.ADDCLIENTLISTENER,
                                                 new Object[]{callbackHandler, callbackHandlerObject}),
                          metadata);
            client.disconnect();
         }
         // now call server to add listener
         invoke(new InternalInvocation(InternalInvocation.ADDLISTENER, null), metadata, clientLocator);
      }
      else
      {
         throw new NullPointerException("InvokerCallbackHandler to be added as a listener can not be null.");
      }
   }

   private Map createListenerMetadata(InvokerCallbackHandler callbackHandler)
   {
      String listenerId = String.valueOf(callbackHandler.hashCode());
      Map metadata = new HashMap();
      metadata.put(LISTENER_ID_KEY, listenerId);
      return metadata;
   }

   /**
    * Removes callback handler as a callback listener from the server (and client in
    * the case that it was setup to receive async callbacks). See addListener().
    *
    * @param callbackHandler
    * @throws Throwable
    */
   public void removeListener(InvokerCallbackHandler callbackHandler) throws Throwable
   {
      if(callbackHandler != null)
      {
         Map metadata = createListenerMetadata(callbackHandler);
         String listenerId = (String) metadata.get(LISTENER_ID_KEY);
         // connect to the given client locator and remove handler as listener
         InvokerLocator locator = invoker.getClientLocator(listenerId);
         if(locator != null) // async callback
         {
            Client client = new Client(locator, subsystem);
            client.setSessionId(getSessionId());
            client.connect();
            client.invoke(new InternalInvocation(InternalInvocation.REMOVECLIENTLISTENER,
                                                 new Object[]{callbackHandler}),
                          metadata);
            client.disconnect();
         }
         // now call server to remove listener
         invoke(new InternalInvocation(InternalInvocation.REMOVELISTENER, null), metadata);
      }
      else
      {
         throw new NullPointerException("Can not remove null InvokerCallbackHandler listener.");
      }
   }

   /**
    * Gets the callbacks for specified callback handler.  The handler is required because an id is generated
    * for each handler.  So if have two callback handlers registered with the same server, no other way to know
    * for which handler to get the callbacks for.
    *
    * @param callbackHandler
    * @return
    * @throws Throwable
    */
   public List getCallbacks(InvokerCallbackHandler callbackHandler) throws Throwable
   {
      if(callbackHandler != null)
      {
         Map metadata = createListenerMetadata(callbackHandler);
         return (List) invoke(new InternalInvocation(InternalInvocation.GETCALLBACKS, null), metadata);
      }
      else
      {
         throw new NullPointerException("Can not remove null InvokerCallbackHandler listener.");
      }
   }

   public void setMarshaller(Marshaller marshaller)
   {
      if(invoker != null && marshaller != null)
      {
         invoker.setMarshaller(marshaller);
      }
   }

   public void setUnMarshaller(UnMarshaller unmarshaller)
   {
      if(invoker != null && unmarshaller != null)
      {
         invoker.setUnMarshaller(unmarshaller);
      }
   }

   /**
    * Takes an inputstream and wraps a server around.  Then calls the target
    * remoting server and passes a proxy for an inputstream to the server's handler.
    * When the server handler calls on this proxy, it will call back on this server
    * wrapped around this inputstream.
    *
    * @param inputStream
    * @param param       invocation payload
    * @return the return value from the invocation
    * @throws Throwable
    */
   public Object invoke(InputStream inputStream, Object param) throws Throwable
   {
      StreamServer streamServer = new StreamServer(inputStream);
      String locator = streamServer.getInvokerLocator();

      // now call on target server and pass locator for stream callbacks
      return invoke(new InternalInvocation(InternalInvocation.ADDSTREAMCALLBACK, new Object[]{locator, param}), null);
   }
}
