/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.remoting;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.Timer;
import java.util.TimerTask;
import org.jboss.remoting.transport.ClientInvoker;

/**
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class ConnectionValidator extends TimerTask
{
   private List listeners = new ArrayList();
   private Client client = null;

   private Timer validatorTimer = null;
   private long pingPeriod = DEFAULT_PING_PERIOD;

   /**
    * Default ping period.  Value is 2 seconds.
    */
   public static final long DEFAULT_PING_PERIOD = 2000;

   public ConnectionValidator(Client client)
   {
      this.client = client;
   }

   private void start()
   {
      validatorTimer = new Timer(false);
      validatorTimer.schedule(this, pingPeriod, pingPeriod);
   }

   private void stop()
   {
      this.cancel();
      validatorTimer.cancel();
      validatorTimer = null;
   }


   public void addConnectionListener(ConnectionListener listener)
   {
      if(listener != null)
      {
         synchronized(listener)
         {
            if(listeners.size() == 0)
            {
               start();
            }
            listeners.add(listener);
         }
      }
   }

   public boolean removeConnectionListener(ConnectionListener listener)
   {
      boolean isRemoved = false;
      if(listener != null)
      {
         synchronized(listeners)
         {
            isRemoved = listeners.remove(listener);
            if(listeners.size() == 0)
            {
               stop();
            }
         }
      }
      return isRemoved;
   }

   /**
    * The action to be performed by this timer task.
    */
   public void run()
   {
      try
      {
         boolean isValid = checkConnection(client.getInvoker());
         if(!isValid)
         {
            notifyListeners(new Exception("Could not connect to server."));
         }
      }
      catch(Throwable thr)
      {
         notifyListeners(thr);
      }
   }

   private void notifyListeners(Throwable thr)
   {
      final Throwable t = thr;
      synchronized(listeners)
      {
         ListIterator itr = listeners.listIterator();
         while(itr.hasNext())
         {
            final ConnectionListener listener = (ConnectionListener) itr.next();
            new Thread()
            {
               public void run()
               {
                  listener.handlerConnectionException(t, client);
               }
            }.start();
         }
      }
   }

   public static boolean checkConnection(ClientInvoker clientInvoker) throws Throwable
   {
      Object o = clientInvoker.invoke(new InvocationRequest(ConnectionValidator.class.getName(),
                                                            Subsystem.SELF, "$PING$", null, null, null));
      InvocationResponse resp = (InvocationResponse) o;
      Boolean b = (Boolean) resp.getResult();
      return b.booleanValue();
   }


}