/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/
package org.jboss.remoting.network;

import javax.management.Notification;
import javax.management.ObjectName;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.detection.ServerInvokerMetadata;
import org.jboss.remoting.ident.Identity;

/**
 * NetworkNotification is a JMX Notification that is sent when changes occur to the network layout as
 * tracked by a NetworkRegistryMBean.
 *
 * @author <a href="mailto:jhaynie@vocalocity.net">Jeff Haynie</a>
 * @version $Revision: 1.4 $
 */
public class NetworkNotification extends Notification
{
   static final long serialVersionUID = 987487760197074685L;

   public static final String SERVER_ADDED = "jboss.network.server.added";
   public static final String SERVER_UPDATED = "jboss.network.server.updated";
   public static final String SERVER_REMOVED = "jboss.network.server.removed";
   public static final String DOMAIN_CHANGED = "jboss.network.domain.changed";

   private final Identity identity;
   private final ServerInvokerMetadata serverInvokers[];
   private final InvokerLocator locators[];


   public NetworkNotification(ObjectName source, String type, Identity identity, ServerInvokerMetadata serverInvokers[])
   {
      super(type, source, System.currentTimeMillis());
      this.identity = identity;
      this.serverInvokers = serverInvokers;
      this.locators = null;
   }

   public NetworkNotification(ObjectName source, String type, Identity identity, InvokerLocator locators[])
   {
      super(type, source, System.currentTimeMillis());
      this.identity = identity;
      this.serverInvokers = null;
      this.locators = locators;
   }

   /**
    * return the identity of the notification
    *
    * @return
    */
   public final Identity getIdentity()
   {
      return identity;
   }

   /**
    * return the locators affected by the notification
    *
    * @return
    */
   public final InvokerLocator[] getLocator()
   {
      if(locators != null)
      {
         return locators;
      }
      else
      {
         InvokerLocator[] locators = new InvokerLocator[serverInvokers.length];
         for(int x = 0; x < serverInvokers.length; x++)
         {
            locators[x] = serverInvokers[x].getInvokerLocator();
         }
         return locators;
      }
   }

   /**
    * Gets all the server invoker metadata for the servers running on
    * specified server.  Will include locator and supported subsystems
    * for that locator.
    *
    * @return
    */
   public final ServerInvokerMetadata[] getServerInvokers()
   {
      return serverInvokers;
   }
}
