/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.remoting.transport.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import org.jboss.logging.Logger;
import org.jboss.remoting.CannotConnectException;
import org.jboss.remoting.ConnectionFailedException;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.RemoteClientInvoker;
import org.jboss.remoting.marshal.Marshaller;
import org.jboss.remoting.marshal.UnMarshaller;
import org.jboss.remoting.marshal.http.HTTPMarshaller;
import org.jboss.util.Base64;

/**
 * HTTP client invoker.  Used for making http requests on http/servlet invoker.
 *
 * @author <a href="mailto:tom@jboss.org">Tom Elrod</a>
 */
public class HTTPClientInvoker extends RemoteClientInvoker
{

   protected final Logger log = Logger.getLogger(getClass());

   public HTTPClientInvoker(InvokerLocator locator)
   {
      super(locator);
   }

   /**
    * @param sessionId
    * @param invocation
    * @param marshaller
    * @return
    * @throws java.io.IOException
    * @throws org.jboss.remoting.ConnectionFailedException
    *
    */
   protected Object transport(String sessionId, Object invocation, Map metadata,
                              Marshaller marshaller, UnMarshaller unmarshaller)
         throws IOException, ConnectionFailedException
   {
      String targetURL = getLocator().getOriginalURI();
      Object httpResponse = useHttpURLConnection(targetURL, invocation, metadata, marshaller, unmarshaller);

      return httpResponse;
   }

   private Object useHttpURLConnection(String url, Object invocation, Map metadata,
                                       Marshaller marshaller, UnMarshaller unmarshaller)
   {
      Object result = null;
      try
      {
         HttpURLConnection conn = createURLConnection(url, metadata);

         // check to see if basic auth required
         String basicAuth = getBasicAuth(metadata);
         if(basicAuth != null)
         {
            conn.setRequestProperty("Authorization", basicAuth);
         }

         // Get the request method type
         boolean isPost = true;
         if(metadata != null)
         {
            String type = (String) metadata.get("TYPE");
            if(type != null && type.equals("GET"))
            {
               isPost = false;
            }

            // Set request headers
            Map header = (Map) metadata.get("HEADER");
            if(header != null)
            {
               Set keys = header.keySet();
               Iterator itr = keys.iterator();
               while(itr.hasNext())
               {
                  String key = (String) itr.next();
                  String value = (String) header.get(key);
                  log.debug("Setting request header with " + key + " : " + value);
                  conn.setRequestProperty(key, value);
               }
            }
         }

         if(isPost)
         {
            //POST
            conn.setDoOutput(true);
            conn.setDoInput(true);
            conn.setRequestMethod("POST");

            OutputStream stream = conn.getOutputStream();
            marshaller.write(invocation, stream);
            InputStream is = conn.getInputStream();
            Map headers = conn.getHeaderFields();
            result = unmarshaller.read(is, headers);
         }
         else
         {
            throw new Exception("HTTP GET opperation not currently supported.");
         }
      }
      catch(Exception e)
      {
         log.debug("Error invoking http client invoker.", e);
         throw new CannotConnectException("Can not connect http client invoker.", e);
      }

      return result;
   }

   protected HttpURLConnection createURLConnection(String url, Map metadata) throws IOException
   {
      URL externalURL = null;
      HttpURLConnection httpURLConn = null;

      // need to find out if need to use a proxy or not
      String proxyHost = null;
      String proxyportString = null;
      int proxyPort = -1;
      boolean proxyOn = true;

      if(metadata != null)
      {
         // first check the metadata as will have precedence
         proxyHost = (String) metadata.get("http.proxyHost");
         proxyportString = (String) metadata.get("http.proxyPort");
         if(proxyportString != null && proxyportString.length() > 0)
         {
            try
            {
               proxyPort = Integer.parseInt(proxyportString);
            }
            catch(NumberFormatException e)
            {
               log.warn("Error converting proxy port specified (" + proxyportString + ") to a number.");
            }
         }
      }

      // if not set in the metadata, try the system properties
      if(proxyHost == null)
      {
         // have found that the actual system property to be set can differ between VMs,
         // so will try all the ones I know of.
         proxyHost = System.getProperty("http.proxyHost");
         if(proxyHost == null)
         {
            proxyHost = System.getProperty("proxyHost");
         }

         // since not set by metadata, need to check if proxy turned off by property setting
         String proxyOnString = System.getProperty("http.proxySet");
         if(proxyOnString == null)
         {
            proxyOnString = System.getProperty("proxySet");
         }
         if(proxyOnString != null)
         {
            proxyOn = Boolean.getBoolean(proxyOnString);
         }
      }
      if(proxyPort < 0)
      {
         String proxyPortString = System.getProperty("http.proxyPort");
         if(proxyPortString == null)
         {
            proxyPortString = System.getProperty("proxyPort");
         }
         if(proxyPortString != null)
         {
            try
            {
               proxyPort = Integer.parseInt(proxyPortString);
            }
            catch(NumberFormatException e)
            {
               log.warn("Error converting proxy port specified (" + proxyportString + ") to a number.");
            }
         }
      }

      // now determin if going to use proxy or not
      if(proxyHost != null && proxyOn)
      {
         // is ok if port is -1 since URL constructor will use default port for protocol
         externalURL = new URL("http", proxyHost, proxyPort, url);

         httpURLConn = (HttpURLConnection) externalURL.openConnection();

         // since know it is a proxy being used, see if have proxy auth
         String proxyAuth = getProxyAuth(metadata);
         if(proxyAuth != null)
         {
            httpURLConn.setRequestProperty("Proxy-Authorization", proxyAuth);
         }
      }
      else
      {
         externalURL = new URL(url);
         httpURLConn = (HttpURLConnection) externalURL.openConnection();
      }

      return httpURLConn;
   }

   private String getProxyAuth(Map metadata)
   {
      String authString = null;
      String username = null;
      String password = null;

      if(metadata != null)
      {
         username = (String) metadata.get("http.proxy.username");
      }
      if(username == null || username.length() == 0)
      {
         username = System.getProperty("http.proxy.username");
      }
      if(metadata != null)
      {
         password = (String) metadata.get("http.proxy.password");
      }
      if(password == null || password.length() == 0)
      {
         password = System.getProperty("http.proxy.password");
      }

      if(username != null && password != null)
      {
         StringBuffer buffer = new StringBuffer();
         buffer.append(username);
         buffer.append(":");
         buffer.append(password);

         String encoded = Base64.encodeBytes(buffer.toString().getBytes());

         authString = "Basic " + encoded;

      }

      return authString;
   }

   private String getBasicAuth(Map metadata)
   {
      String authString = null;
      String username = null;
      String password = null;

      if(metadata != null)
      {
         username = (String) metadata.get("http.basic.username");
      }
      if(username == null || username.length() == 0)
      {
         username = System.getProperty("http.basic.username");
      }
      if(metadata != null)
      {
         password = (String) metadata.get("http.basic.password");
      }
      if(password == null || password.length() == 0)
      {
         password = System.getProperty("http.basic.password");
      }

      if(username != null && password != null)
      {
         StringBuffer buffer = new StringBuffer();
         buffer.append(username);
         buffer.append(":");
         buffer.append(password);

         String encoded = Base64.encodeBytes(buffer.toString().getBytes());

         authString = "Basic " + encoded;

      }

      return authString;
   }


   /**
    * subclasses must implement this method to provide a hook to connect to the remote server, if this applies
    * to the specific transport. However, in some transport implementations, this may not make must difference since
    * the connection is not persistent among invocations, such as SOAP.  In these cases, the method should
    * silently return without any processing.
    *
    * @throws org.jboss.remoting.ConnectionFailedException
    *
    */
   protected void handleConnect() throws ConnectionFailedException
   {
      // NO OP as not statefull connection
   }

   /**
    * subclasses must implement this method to provide a hook to disconnect from the remote server, if this applies
    * to the specific transport. However, in some transport implementations, this may not make must difference since
    * the connection is not persistent among invocations, such as SOAP.  In these cases, the method should
    * silently return without any processing.
    */
   protected void handleDisconnect()
   {
      // NO OP as not statefull connection
   }

   /**
    * Each implementation of the remote client invoker should have
    * a default data type that is uses in the case it is not specified
    * in the invoker locator uri.
    *
    * @return
    */
   protected String getDefaultDataType()
   {
      return HTTPMarshaller.DATATYPE;
   }

}