/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/
package org.jboss.remoting.transport.http;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;
import javax.management.MBeanServer;
import javax.management.MBeanServerInvocationHandler;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpParser;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.marshal.MarshalFactory;
import org.jboss.remoting.marshal.Marshaller;
import org.jboss.remoting.marshal.UnMarshaller;
import org.jboss.remoting.marshal.http.HTTPMarshaller;
import org.jboss.remoting.marshal.http.HTTPUnMarshaller;
import org.jboss.remoting.marshal.serializable.SerializableMarshaller;
import org.jboss.remoting.transport.web.WebServerInvoker;
import org.jboss.util.threadpool.BasicThreadPool;
import org.jboss.util.threadpool.BlockingMode;
import org.jboss.util.threadpool.ThreadPool;
import org.jboss.util.threadpool.ThreadPoolMBean;

/**
 * This is the stand alone http server invoker which acts basically as a web server.
 *
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 */

/**
 * Server invoker implementation based on http protocol.  Is basically a stand alone http server whose request are
 * forwared to the invocation handler and responses from invocation handler are sent back to caller as http response.
 */
public class HTTPServerInvoker extends WebServerInvoker implements Runnable
{
   public static final String MAX_NUM_HTTP_THREADS_KEY = "maxNumThreadsHTTP";
   public static final String HTTP_THREAD_POOL_CLASS_KEY = "HTTPThreadPool";

   private String httpThreadPoolClass = null;

   private static int BACKLOG_DEFAULT = 1000;
   private static int MAX_POOL_SIZE_DEFAULT = 100;

   private ServerSocket serverSocket = null;

   private boolean running = false;

   private ThreadPool httpThreadPool;
   private int maxPoolSize = MAX_POOL_SIZE_DEFAULT;

   protected int backlog = BACKLOG_DEFAULT;

   // list of content types
   public static String HTML = "text/html";
   public static String PLAIN = "text/plain";
   public static String SOAP = "application/soap+xml";

   public HTTPServerInvoker(InvokerLocator locator)
   {
      super(locator);
   }

   public HTTPServerInvoker(InvokerLocator locator, Map configuration)
   {
      super(locator, configuration);
   }

   protected String getDefaultDataType()
   {
      return SerializableMarshaller.DATATYPE;
   }

   protected void setup() throws Exception
   {
      super.setup();

      Map config = getConfiguration();
      String maxNumOfThreads = (String) config.get(MAX_NUM_HTTP_THREADS_KEY);
      if(maxNumOfThreads != null && maxNumOfThreads.length() > 0)
      {
         try
         {
            maxPoolSize = Integer.parseInt(maxNumOfThreads);
         }
         catch(NumberFormatException e)
         {
            log.error("Can not convert max number of threads value (" + maxNumOfThreads + ") into a number.");
         }
      }
      httpThreadPoolClass = (String) config.get(HTTP_THREAD_POOL_CLASS_KEY);

   }

   public void setMaxNumberOfHTTPThreads(int numOfThreads)
   {
      this.maxPoolSize = numOfThreads;
   }

   public int getMaxNumberOfHTTPThreads()
   {
      return this.maxPoolSize;
   }

   public ThreadPool getHTTPThreadPool()
   {
      if(httpThreadPool == null)
      {
         // if no thread pool class set, then use default BasicThreadPool
         if(httpThreadPoolClass == null || httpThreadPoolClass.length() == 0)
         {
            BasicThreadPool basicthreadpool = new BasicThreadPool("JBossRemoting - HTTP Server Invoker");
            basicthreadpool.setBlockingMode(BlockingMode.RUN);
            basicthreadpool.setMaximumPoolSize(maxPoolSize);
            httpThreadPool = basicthreadpool;
         }
         else
         {
            //first check to see if this is an ObjectName
            boolean isObjName = false;
            try
            {
               ObjectName objName = new ObjectName(httpThreadPoolClass);
               httpThreadPool = createThreadPoolProxy(objName);
               isObjName = true;
            }
            catch(MalformedObjectNameException e)
            {
               log.debug("Thread pool class supplied is not an object name.");
            }

            if(!isObjName)
            {
               try
               {
                  httpThreadPool = (ThreadPool) getClassLoader().loadClass(httpThreadPoolClass).newInstance();
               }
               catch(Exception e)
               {
                  throw new RuntimeException("Error loading instance of ThreadPool based on class name: " + httpThreadPoolClass);
               }
            }
         }
      }
      return httpThreadPool;
   }

   private ThreadPool createThreadPoolProxy(ObjectName objName)
   {
      ThreadPool pool;
      MBeanServer server = getMBeanServer();
      if(server != null)
      {
         ThreadPoolMBean poolMBean = (ThreadPoolMBean)
               MBeanServerInvocationHandler.newProxyInstance(server,
                                                             objName,
                                                             ThreadPoolMBean.class,
                                                             false);
         pool = poolMBean.getInstance();
      }
      else
      {
         throw new RuntimeException("Can not register MBean ThreadPool as the ServerInvoker has not been registered with a MBeanServer.");
      }
      return pool;
   }


   public void setHTTPThreadPool(ThreadPool pool)
   {
      this.httpThreadPool = pool;
   }


   public void start() throws IOException
   {
      if(!running)
      {
         running = true;

         try
         {
            ThreadPool httpThreadPool = getHTTPThreadPool();
            InetAddress bindAddress = InetAddress.getByName(getServerBindAddress());
            serverSocket = createServerSocket(getServerBindPort(), backlog, bindAddress);

            // prime the pool by starting up max
            for(int t = 0; t < maxPoolSize; t++)
            {
               httpThreadPool.run(this);
            }
         }
         catch(IOException e)
         {
            log.error("Error starting ServerSocket.  Bind port: " + getServerBindPort() + ", bind address: " + getServerBindAddress());
            throw e;
         }
      }
      super.start();
   }

   protected ServerSocket createServerSocket(int serverBindPort, int backlog, InetAddress bindAddress) throws IOException
   {
      return new ServerSocket(serverBindPort, backlog, bindAddress);
   }

   public void run()
   {
      try
      {
         Socket socket = serverSocket.accept();

         if(socket != null)
         {
            httpThreadPool.run(this);
            processRequest(socket);
         }
      }
      catch(Throwable thr)
      {
         if(running)
         {
            log.error("Error processing incoming request.", thr);
         }
      }
   }

   public void stop()
   {
      if(running)
      {
         running = false;

         maxPoolSize = 0; // so ServerThreads don't reinsert themselves

         try
         {
            httpThreadPool.stop(false);
            httpThreadPool.waitForTasks(2000);
         }
         catch(InterruptedException e)
         {
            e.printStackTrace();
         }

         try
         {
            if(serverSocket != null && !serverSocket.isClosed())
            {
               serverSocket.close();
            }
            serverSocket = null;
         }
         catch(Exception e)
         {
         }
      }
      super.stop();

      log.debug("HTTPServerInvoker stopped.");
   }

   private void processRequest(Socket socket)
   {
      DataInputStream dataInput = null;
      DataOutputStream dataOutput = null;
      BufferedOutputStream bufferOutput = null;
      try
      {
         Object response = null;
         boolean isError = false;
         String requestContentType = null;
         try
         {
            dataInput = new DataInputStream(socket.getInputStream());

            // Need to parse the header to find Content-Type

            /**
             * Read the first line, as this will be the POST or GET, path, and HTTP version.
             * Then next comes the headers.  (key value seperated by a ': '
             * Then followed by an empty \r\n, which will be followed by the payload
             */
            ByteArrayOutputStream buffer = new ByteArrayOutputStream();
            int ch;
            while((ch = dataInput.read()) >= 0)
            {
               buffer.write(ch);
               if(ch == '\n')
               {
                  break;
               }
            }

            String methodType = null;
            String path = null;
            String httpVersion = null;

            byte[] firstLineRaw = buffer.toByteArray();
            // now need to backup and make sure the character before the \n was a \r
            if(firstLineRaw[firstLineRaw.length - 2] == '\r')
            {
               //Got our first line, now to set the variables
               String firstLine = new String(firstLineRaw).trim();
               int startIndex = 0;
               int endIndex = firstLine.indexOf(' ');
               methodType = firstLine.substring(startIndex, endIndex);
               startIndex = endIndex + 1;
               endIndex = firstLine.indexOf(' ', startIndex);
               path = firstLine.substring(startIndex, endIndex);
               startIndex = endIndex + 1;
               httpVersion = firstLine.substring(startIndex);
            }
            else
            {
               log.error("Error processing first line.  Should have ended in \r\n, but did not");
               throw new RuntimeException("Error processing HTTP request type.  First line of request is invalid.");
            }

            Map metadata = new HashMap();
            Header[] headers = HttpParser.parseHeaders(dataInput);
            for(int x = 0; x < headers.length; x++)
            {
               String headerName = headers[x].getName();
               String headerValue = headers[x].getValue();
               metadata.put(headerName, headerValue);
               // doing this instead of getting from map since ignores case
               if("Content-Type".equalsIgnoreCase(headerName))
               {
                  requestContentType = headers[x].getValue();
               }
            }

            UnMarshaller unmarshaller = MarshalFactory.getUnMarshaller(HTTPUnMarshaller.DATATYPE);
            Object obj = unmarshaller.read(dataInput, metadata);

            InvocationRequest request = null;

            if(obj instanceof InvocationRequest)
            {
               request = (InvocationRequest) obj;
            }
            else
            {
               if(isBinary(requestContentType))
               {
                  request = getInvocationRequest(metadata, obj);
               }
               else
               {
                  request = createNewInvocationRequest(metadata, obj);
               }
            }

            try
            {
               // call transport on the subclass, get the result to handback
               response = invoke(request);
            }
            catch(Throwable ex)
            {
               log.debug("Error thrown calling invoke on server invoker.", ex);
               response = ex;
               isError = true;
            }

            dataOutput = new DataOutputStream(socket.getOutputStream());
            bufferOutput = new BufferedOutputStream(dataOutput, 1024);
         }
         catch(Throwable thr)
         {
            log.debug("Error thrown processing request.  Probably error with processing headers.", thr);
            if(thr instanceof Exception)
            {
               response = (Exception) thr;
            }
            else
            {
               response = new Exception(thr);
            }
            isError = true;
         }
         if(bufferOutput != null)
         {
            if(response == null)
            {
               //TODO: -TME Shoudl this be 1.0 or 1.1?
               bufferOutput.write("HTTP/1.1 ".getBytes());
               bufferOutput.write("200 JBoss Remoting: Request Processed Successfully".getBytes());
               String contentType = "\r\n" + "Content-Type" + ": " + "text/html";
               bufferOutput.write(contentType.getBytes());
               String contentLength = "\r\n" + "Content-Length" + ": " + 0;
               bufferOutput.write(contentLength.getBytes());
               //dataOutput.write(("\r\n" + "\r\n").getBytes());
               //dataOutput.write("NULL return".getBytes());
               bufferOutput.flush();
            }
            else
            {
               //TODO: -TME Shoudl this be 1.0 or 1.1?
               bufferOutput.write("HTTP/1.1 ".getBytes());
               String status = isError ? "500 JBoss Remoting: Error occurred within target application." :
                               "200 JBoss Remoting: Request Processed Succussfully";
               bufferOutput.write(status.getBytes());

               // write return headers
               String contentType = "\r\n" + "Content-Type" + ": " + requestContentType;
               bufferOutput.write(contentType.getBytes());
               int iContentLength = getContentLength(response);
               String contentLength = "\r\n" + "Content-Length" + ": " + iContentLength;
               bufferOutput.write(contentLength.getBytes());
               // content seperator
               bufferOutput.write(("\r\n" + "\r\n").getBytes());

               // write response
               //Marshaller marshaller = getMarshallerByContentType(requestContentType);
               Marshaller marshaller = MarshalFactory.getMarshaller(getLocator(), this.getClass().getClassLoader());
               if(marshaller == null)
               {
                  marshaller = MarshalFactory.getMarshaller(HTTPMarshaller.DATATYPE);
               }
               marshaller.write(response, bufferOutput);

            }
         }
         else
         {
            if(isError)
            {
               log.warn("Can not send error response due to output stream being null (due to previous error).");
            }
            else
            {
               log.error("Can not send response due to output stream being null (even though there was not a previous error encountered).");
            }
         }
      }
      catch(Exception e)
      {
         log.error("Error processing client request.", e);
      }
      finally
      {
         if(dataInput != null)
         {
            try
            {
               dataInput.close();
            }
            catch(Exception e)
            {
               log.warn("Error closing resource.", e);
            }
         }
         if(dataOutput != null)
         {
            try
            {
               dataOutput.close();
            }
            catch(Exception e)
            {
               log.warn("Error closing resource.", e);
            }
         }
         try
         {
            socket.close();
         }
         catch(Exception e)
         {
            log.warn("Error closing resource.", e);
         }
      }
   }
}
