/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * 
 * Created on Aug 4, 2005
 *
 */
package org.jboss.remoting.transport.multiplex;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;

import org.jboss.logging.Logger;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentHashMap;

/**
 * 
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class InputMultiplexor
{
   protected static final Logger log = Logger.getLogger(InputMultiplexor.class);
   
   
   /**
    * 
    */
   public InputMultiplexor()
   {
   }

   

/**
 *
 * @return
 */
   static InputThread getAnInputThread(MultiplexingManager manager, InputStream is, OutputStream os)
   {
      return new InputThread(manager, is, os);
   }



   public static class InputThread extends StoppableThread
   {
      private static final Logger log = Logger.getLogger(InputMultiplexor.InputThread.class);
      
      private static final int READING_DATA = 0;
      private static final int SAW_MARKER = 1;
      private static final int READING_ID = 2;
      
      private InputStream is;
      private OutputStream currentOutputStream;
      private ByteArrayOutputStream bos;
      private MultiplexingManager manager;
      private int state;
      private boolean managerShuttingDown = false;
      private int markerCount = 0;
      private int dataInCount = 0;
      private boolean trace;
      
      
      public InputThread(MultiplexingManager manager, InputStream is, OutputStream os)
      {
         this.is = is;
         this.manager = manager;
         
         currentOutputStream = os;
         bos = new ByteArrayOutputStream();
         int state = READING_DATA;
         trace = log.isTraceEnabled();
      }
      
      
      public synchronized void shutdown()
      {
         // in case thread is still reading
         super.shutdown();
         
//         try
//         {
//            is.close();
//         }
//         catch (IOException e)
//         {
//            log.debug("input thread: i/o exception while shutting down");
//         }
         
         interrupt();
         
         // in case thread is in stayAlive()
         managerShuttingDown = true;
         notify();
      }
      
      
      /**
       * 
       */
      protected void doInit()
      {
         log.info("input thread starting");
      }
      
      
      /**
       * 
       */
      protected void doRun()
      {
         int b = 0;
         
         try
         {
            b = is.read();
            
            while (b >= 0 && isRunning())
            {
               if (b != OutputMultiplexor.MARKER)
               {
                  if (markerCount > 0)
                     log.debug("read " + markerCount + " MARKER's");
                  
                  markerCount = 0;
                  log.debug("reading int: " +  (int) b);
               }
               
               switch(state)
               {
               
               case READING_DATA:
                  
                  if (b == OutputMultiplexor.MARKER)
                     state = SAW_MARKER;
                  else
                  {
                     currentOutputStream.write(b);
                     dataInCount++;
                  }
                  
                  break;
                  
                  
               case SAW_MARKER:
                  
                  if (b == OutputMultiplexor.MARKER)
                  {
                     state = READING_DATA;
                     currentOutputStream.write(b);
                     markerCount++;
                     dataInCount++;
                  }
                  else
                  {
                     state = READING_ID;
                     currentOutputStream.flush();
                     bos.reset();
                     bos.write(b);
                     
                     if (markerCount > 0) log.debug("read " + markerCount + " MARKER's");
                     markerCount = 0;
                  }
                  
                  break;
                  
               case READING_ID:
                  
                  if (b == OutputMultiplexor.MARKER)
                  {
                     state = READING_DATA;
                     SocketId socketId = new SocketId(bos.toByteArray());
                     bos.reset();
                     currentOutputStream = manager.getOutputStreamByLocalSocket(socketId);
                     log.debug("redirecting input to socket: " + socketId.getPort());
                     
                     if (currentOutputStream == null)
                     {
                        log.error("unknown socket id: " + socketId.getPort());
                        currentOutputStream = manager.getOutputStreamByLocalSocket(SocketId.DEADLETTER_SOCKET_ID);
                        
                        if (currentOutputStream == null)
                        {
                           log.error("cannot find dead letter output stream");
                           currentOutputStream = new ByteArrayOutputStream();
                        }
                     } 
                  }
                  else
                     bos.write(b);
                  
                  break; 
                  
               } // switch
               
               b = is.read();
               
            } // while
            
         } // try
         
         catch (InterruptedIOException e)
         {
            handleError("interrupted", e);
         }
         catch (IOException e)
         {
            handleError("i/o error", e);
         }
         
         super.shutdown();
         
         log.debug("last byte: " + b);
         log.debug("isRunning: " + isRunning());
         log.debug("read " + markerCount + " MARKER's");
      }
      
      
      protected void doShutDown()
      {
         //            try
         //            {
         //               is.close();
         //            }
         //            catch (Exception ignored)
         //            {
         //               if (trace)
         //                  log.trace("unable to close InputStream: " + ignored.getMessage(), ignored);
         //            }
         
         //            TODO: work this out
         //            if (terminatedOnError)
         //               multiplexor.shutdown();
         
         log.info("input thread: data bytes read: " + dataInCount);
         
         flushPipes();
         stayAlive();
         
         log.info("input thread shutting down");
      }
      
      
      /**
       *
       * @param error
       * @param e
       */
      protected void handleError(String error, Throwable e)
      {
         if (e instanceof InterruptedException || e instanceof IOException)
         {
            if (trace)
               log.trace(error, e);
         }
         else
            log.debug(error, e);
         
         //    TODO
         //            running = false;
         //            terminatedOnError = true;
      }
      
      
      protected synchronized void stayAlive()
      {
         log.info("input thread entering stayAlive()");
         
         while (!managerShuttingDown)
         {
            try
            {
               wait();
            }
            catch (InterruptedException e)
            {
               log.debug("stayAlive(): InterruptedException");
            }
            
            log.debug("stayAlive(): wait ended");
         }
      }
      
      
      protected void flushPipes()
      {
         Iterator it = manager.getAllOutputStreams().iterator();
         
         while (it.hasNext())
         {
            OutputStream os = (OutputStream) it.next();
            
            try
            {
               os.flush();
//             os.close();
            }
            catch (IOException ignored)
            {
               log.error("unable to flush a PipedOutputStream");
            }
         }
         
         log.info("input thread flushed PipedOutputStream's");
      }
   }
}
