/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * 
 * Created on Jul 31, 2005
 */
package org.jboss.remoting.transport.multiplex;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.SocketException;

import org.jboss.logging.Logger;

/**
 * 
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class MultiplexingInputStream extends PipedInputStream
{
   protected static final Logger log = Logger.getLogger(MultiplexingInputStream.class);
   private VirtualSocket socket;
   private PipedOutputStream sourceStream;
   private boolean eof = false;
   private boolean closed = false;
   private boolean remoteShutDownPending = false;
   private Thread readThread;

   
/**
 * 
 */
   public MultiplexingInputStream(PipedOutputStream sourceStream, VirtualSocket socket) throws IOException
   {
      super(sourceStream);
      this.sourceStream = sourceStream;
      this.socket = socket;
   }
   
   
/**
 * @param src
 * @throws java.io.IOException
 */
   public MultiplexingInputStream(PipedOutputStream sourceStream) throws IOException
   {
      super(sourceStream);
      this.sourceStream = sourceStream;
   }
      
   
/**
 * 
 */
   public int read() throws IOException
   {
      /*
       * It would be nice to call MultiplexingInputStream.read(byte[], int, int) to avoid code
       * duplication.  However, MultiplexingInputStream.read(byte[], int, int) calls
       * super.read(byte[], int, int), which calls read(), creating a loop.
       */
      
      if (eof)
         return -1;
      
      if (closed)
         throw new SocketException("Socket closed");
      
      if (remoteShutDownPending)
         if (available() == 0)
         {
            remoteShutDownPending = false;
            setEOF();
            return -1;
         }

      try
      {
         synchronized(this)
         {
            readThread = Thread.currentThread();
            log.debug("read(): calling super.read()");
            int b = super.read();
            readThread = null;
            log.debug("read(): returning from super.read()");
            return b;
         }
      }
      catch (InterruptedIOException e)
      {
         if (remoteShutDownPending)
         {
            remoteShutDownPending = false;
            setEOF();
            return -1;
         }
         
         throw e;
      }
      catch (IOException e)
      {
         if (closed)
         {
            throw new SocketException("Socket closed");
         }
         else if (remoteShutDownPending)
         {  // if notified of remote output shutdown during super.read()
            log.debug("read(): interrupted due to remote shutdown pending");
            remoteShutDownPending = false;
            setEOF();
            return -1;
         }
         else
            throw e;
      }
   }
   
   
/**
 * 
 */
   public int read(byte[] bytes) throws IOException
   {
      return read(bytes, 0, bytes.length);
   }
   
   
/**
 * 
 */
   public int read(byte[] bytes, int off, int len) throws IOException
   {
      if (eof)
         return -1;
      
      if (closed)
         throw new SocketException("Socket closed");
      
      if (remoteShutDownPending)
         if (available() == 0)
         {
            remoteShutDownPending = false;
            setEOF();
            return -1;
         }
         
      try
      {
         // We leave a reference to the current thread so that handleRemoteShutdown() can
         // interrupt it if necessary.
         synchronized(this)
         {
            readThread = Thread.currentThread();
            log.debug("read(): calling super.read()");
            int n = super.read(bytes, off, len);
            readThread = null;
            log.debug("read(): returning from super.read()");
            return n;
         }
      }
      catch (InterruptedIOException e)
      {
         if (remoteShutDownPending)
         {
            remoteShutDownPending = false;
            setEOF();
            return -1;
         }
         
         throw e;
      }
      catch (IOException e)
      {     
         if (closed)
         {
            throw new SocketException("Socket closed");
         }
         else if (remoteShutDownPending)
         {  // if notified of remote output shutdown during super.read()
            log.debug("read(): interrupted due to remote shutdown pending");
            remoteShutDownPending = false;
            setEOF();
            return -1;
         }
         else
            throw e;
      }
   }
   
   
/**
 * 
 */
   public void close() throws IOException
   {
      if (closed)
         return;
      
      closed = true;
      super.close();
      
      if (socket != null)
         socket.close();
   }

   
/**
 *  FIXME Comment this
 * 
 *
 */
   protected void handleRemoteShutdown() throws IOException
   {
      log.debug("entering handleRemoteShutdown()");
      
      if (eof)
         return;
      
      remoteShutDownPending = true;  
     
      if (available() == 0)
      {  
         // Hack Alert.  If a thread is blocked in read(), we need to interrupt it.
         // This works because when PipedInputStream has no bytes to read, it loops in a wait(). 
         synchronized(this)
         {
            if (readThread != null)
               readThread.interrupt();
         }
      }
      
      //temporary
      //super.close();
   }


/**
 */
   protected void setEOF()
   {
      eof = true;
   }
   
   
   class IOExceptionWrapper
   {
      private IOException e;
      IOException get() {return e;}
      void set(IOException e) {this.e = e;}
      boolean isSet() {return e != null;}
   }
   
   class IntWrapper
   {
      private int i;
      private boolean set = false;
      int get() {return i;}
      void set(int i) {this.i = i; set = true;}
      boolean isSet() {return set;}
   }
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   protected int timedRead() throws IOException
   {
      final IntWrapper b = new IntWrapper();
      final IOExceptionWrapper readException = new IOExceptionWrapper();
      final Thread thisThread = Thread.currentThread();
      final Object lock = new Object();
      
      final Thread readThread = new Thread()
                          {
                             public void run()
                             {
                                try
                                {
                                   synchronized(lock)
                                   {
                                      log.info("readThread: entered sync section");
                                      b.set(MultiplexingInputStream.super.read());
                                      lock.notifyAll();
                                   }
                                }
                                catch (IOException e)
                                {
                                   readException.set(e);
                                }
                             }
                          };
      
      Thread waitThread = new Thread()
                          {
              	             public void run()
              	             {
              	                synchronized(lock)
              	                {
              	                   readThread.start();
              	                   lock.notifyAll();
              	                   try
              	                   {
              	                      lock.wait();
              	                   }
              	                   catch (InterruptedException e)
              	                   {
              	                      log.info(e);
              	                   }
              	                   
              	                   thisThread.interrupt();
              	                }
              	             }
                          };
                          
      waitThread.start();
      
      while(!b.isSet() && !readException.isSet() && !remoteShutDownPending && !eof)
      {
         try
         {
            Thread.sleep(2000);
            log.debug("timedRead(): waiting 1000");
         }
         catch (InterruptedException e)
         {
            log.info("timeRead(): interrupted");
         }
      }
      
      if (readException.isSet())
         throw readException.get();
      
      log.info("timedRead(): returning: " + b.get());
      return b.get();
   }
}
