/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 * Created on Jul 22, 2005
 */
package org.jboss.remoting.transport.multiplex;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PipedOutputStream;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.jboss.logging.Logger;


/**
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class MultiplexingManager
{
   private static final Logger log = Logger.getLogger(MultiplexingManager.class);
   
   /** A HashMap<InetSocketAddress, MultiplexingManager> of MultiplexingManager's indexed by local address */
   // private static ConcurrentHashMap managersBySocketAddress = new ConcurrentHashMap();
   private static Map managersByLocalAddress = Collections.synchronizedMap(new HashMap());
 
   /** A HashMap<InetSocketAddress, MultiplexingManager> of MultiplexingManager's indexed by remote address */
//   private static ConcurrentHashMap managersBySocketAddress = new ConcurrentHashMap();
   private static Map managersByRemoteAddress = Collections.synchronizedMap(new HashMap());
   
   /** OutputMultiplexor associated with this MultiplexingManager */
   private OutputMultiplexor outputMultiplexor;
   
   /** A HashMap<SocketId, VirtualSocket> of VirtualSocket's indexed by local SocketId */
   private Map socketMap = Collections.synchronizedMap(new HashMap());
   
   /** A HashSet of remote SocketId's to which this MultiplexingManager's VirtualSocket's are connected */
   private Set connectedRemoteSockets = Collections.synchronizedSet(new HashSet());
   
   /** A HashMap<Long, InputStream> of InputStream's associated with virtual sockets  */
   private Map outputStreamMap = Collections.synchronizedMap(new HashMap());

   /** Protocol object for handling connection/disconnection communications */
   private Protocol protocol;
   
   /** Actual local socket upon which this family of virtual sockets is based */
   private Socket socket;
   
   /** bound state of actual local socket */
   private boolean bound = false;
   
   /** connected state of actual local socket */
   private boolean connected = false;
   
   /** SocketAddress of remote actual socket to which this Manager's socket is connected */
   private InetSocketAddress remoteSocketAddress;
   
   /** SocketAddress to which this manager's actual socket is bound */
   private InetSocketAddress localSocketAddress;
   
   /** Represents local port on which this manager's actual socket is bound, with any local address */
   private InetSocketAddress localWildCardAddress;
   
   /** Currently registered server socket */
   private ServerSocket serverSocket;
   
   /** Indicates if remote server socket has been registered */
   private boolean remoteServerSocketRegistered = false;
   
   /** InputThread reading from socket's InputStream and distributing to virtual sockets */
   private InputMultiplexor.InputThread inputThread;
   
   /** Thread for writing to socket's OutputStream */
   private OutputMultiplexor.OutputThread outputThread;
   
   /** Thread for getting asynchronous messages from remote Protocol */
   private Protocol.BackChannelThread backChannelThread;
   
   /** OutputStream for unknown virtual sockets */
   private OutputStream deadLetterOutputStream = new ByteArrayOutputStream();
   
   /** Number of VirtualSockets using this MultiplexingManager */
   private int referenceCount = 0;
   
   public static final boolean ON_CLIENT = true;
   
   public static final boolean  ON_SERVER = false;
   
   // TODO: work out isOnClient
   private static boolean  isOnClient = ON_SERVER;
   
   private boolean running;
   
   /** The logging trace level */
   private boolean trace;
   
   private long id;
   

   
   
///**
// * 
// * @param localPort
// */
//   protected MultiplexingManager(int isOnClient, int localPort) throws IOException
//   {
//      socket = new Socket();
//      localSocketAddress = new InetSocketAddress(InetAddress.getLocalHost(), localPort);
//      socket.bind(localSocketAddress);
//      registerBinding();
//   }
//   
//   
///**
// * 
// * @param localPort
// */
//   protected MultiplexingManager(int localPort) throws IOException
//   {
//      this(ON_CLIENT, localPort);
//   }
   
   
   protected MultiplexingManager()
   {
      id = new Date().getTime();
      socket = new Socket();
      
      try
      {
         socket.setReuseAddress(true);
      }
      catch (SocketException ignored)
      {
         log.error(ignored);
      }
   }

   
/**
 * 
 * @param socket
 * @throws IOException
 */
   protected MultiplexingManager(Socket socket) throws IOException
   {
      this.socket = socket;
      id = new Date().getTime();
      setup();
   }
   
   
/**
 * 
 */
   protected MultiplexingManager(String host, int port) throws UnknownHostException, IOException
   {
      socket = new Socket();
      id = new Date().getTime();
     
      try
      {
         socket.setReuseAddress(true);
      }
      catch (SocketException ignored)
      {
         log.error(ignored);
      }
      
      socket.connect(new InetSocketAddress(host, port));
      setup();
   } 
   
   
/**
 * 
 * @param address
 * @throws IOException
 */
   protected MultiplexingManager(InetSocketAddress address) throws IOException
   {
//      socket = new Socket(address.getAddress(), address.getPort());
      socket = new Socket();
      
      try
      {
         socket.setReuseAddress(true);
      }
      catch (SocketException ignored)
      {
         log.error(ignored);
      }
      
      socket.connect(address);
      setup();
   }
      
       
/**
 *
 */
   protected synchronized void setup() throws IOException
   {      
      // register dead letter output stream (for unrecognized destinations)
      outputStreamMap.put(SocketId.DEADLETTER_SOCKET_ID, deadLetterOutputStream);
      
      // TODO: what was this for???
      connectedRemoteSockets.add(SocketId.PROTOCOL_SOCKET_ID);
      connectedRemoteSockets.add(SocketId.SERVER_SOCKET_ID);
      connectedRemoteSockets.add(SocketId.BACKCHANNEL_SOCKET_ID);
      
      // start output thread
      outputMultiplexor = new OutputMultiplexor(this, socket.getOutputStream());
      outputThread = outputMultiplexor.getAnOutputThread();
      outputThread.start();
      
      // start input thread
      inputThread = InputMultiplexor.getAnInputThread(this, socket.getInputStream(), deadLetterOutputStream);
      inputThread.start();      

      // start back channel thread
      protocol = new Protocol(this);
      backChannelThread = protocol.getBackChannelThread();
      backChannelThread.start();
      
      if (!bound)
      {
         bound = true;
         registerByLocalAddress(new InetSocketAddress(socket.getLocalAddress(), socket.getLocalPort()));
      }
      
      if (!connected)
      {
         connected = true;
         registerByRemoteAddress(new InetSocketAddress(socket.getInetAddress(), socket.getPort()));
      }
      
//      // register this MultiplexorManager by local actual socket address
//      localSocketAddress = new InetSocketAddress(socket.getLocalPort());
//      managersBySocketAddress.put(localSocketAddress, this);
//      
//      // register this MultiplexorManager by remote actual socket address
//      // TODO: work out unconnected socket
//      remoteSocketAddress = new InetSocketAddress(socket.getInetAddress(), socket.getPort());
//      managersBySocketAddress.put(remoteSocketAddress, this);
      
      trace = log.isTraceEnabled();     
   }
   
   
/**
 * 
 * @param port
 * @return
 * @throws IOException
 */
   public static MultiplexingManager getaManager(int port) throws IOException
   {
      log.debug("entering getaManager(int port)");
      return getaManagerByLocalAddress(new InetSocketAddress(port));
   }
   
   
/**
 * 
 * @param socket
 * @return
 * @throws IOException
 * 
 * TODO: what if multiplexor already exists?
 */
   public static MultiplexingManager getaManager(Socket socket) throws IOException
   {
      log.debug("entering getaManager(Socket socket");
      return new MultiplexingManager(socket);
   }
   

/**
 * 
 * @param address
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager getaManagerByLocalAddress(InetSocketAddress address) throws IOException
   {
      log.debug("entering getaManagerByLocalAddress(InetSocketAddress address)");
      MultiplexingManager  m = (MultiplexingManager) managersByLocalAddress.get(address);
      
      if (m != null)
      {
         return m;
      }
      
      m = new MultiplexingManager();
      m.bind(address);
      return m;
   }  
   
/**
 * 
 * @param address
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager getaManagerByRemoteAddress(InetSocketAddress address) throws IOException
   {
      log.debug("entering getaManagerByRemoteAddress(InetSocketAddress address)");
      MultiplexingManager  m = (MultiplexingManager) managersByRemoteAddress.get(address);
      
      try
      {
         if (m != null && m.getProtocol().isRemoteServerSocketRegistered())
         {
            return m;
         }
      }
      catch (Exception e)
      {
         log.info(m.getSocket().toString() + ": unable to call isRemoteServerSocketRegistered(): shutting down");
         m.shutdown();
      }
      
      return new MultiplexingManager(address);
   }
   
   
/**
 * 
 * @return
 */
   public static boolean isOnClient()
   {
      return isOnClient;
   }
   
   
   public static void setOnClient(boolean isOnClient)
   {
      log.info("setting onClient: " + isOnClient);
      MultiplexingManager.isOnClient = isOnClient;
   }
      
      
/**
 *
 */
   public synchronized void incrementReferences()
   {
      referenceCount++;
      log.info(socket.toString() + ": incrementReferences(): number of references: " + referenceCount);
   }
     
     
/**
 *  FIXME Comment this
 * 
 *
 */
   public synchronized void releaseReference()
   {
      referenceCount--;
      log.info(socket.toString() + "releaseReference(): number of references: " + referenceCount);
      
      if (referenceCount == 0)
      {
         // unregister this MultiplexorManager by local actual socket address
         managersByLocalAddress.remove(localSocketAddress);
         log.info(socket.toString() + ": deleting local reference: " + localSocketAddress);
         
         // unregister this MultiplexorManager by wildcard version of local actual socket address
         managersByLocalAddress.remove(localWildCardAddress);
         log.info(socket.toString() + ": deleting local reference: " + localWildCardAddress);
         
         // unregister this MultiplexorManager by remote actual socket address
         managersByRemoteAddress.remove(remoteSocketAddress);
         log.info(socket.toString() + ": deleting remote reference: " + remoteSocketAddress);
         
         shutdown();
      }
   }
   
   
/*
 * 
 */
   public synchronized int getReferenceCount()
   {
      return referenceCount;
   }
   
   
/**
 * 
 * @param address
 * @throws IOException
 */
   public synchronized void bind(InetSocketAddress address) throws IOException
   {
      if (bound)
         throw new IOException("socket is already bound");
      
      socket.bind(address);
      registerByLocalAddress(address);
      bound = true;
   }
   
   
   protected synchronized void registerByLocalAddress(InetSocketAddress address)
   {
      localSocketAddress = address;
      managersByLocalAddress.put(address, this);
      
      // allow searching on any local address
      localWildCardAddress = new InetSocketAddress(address.getPort());
      managersByLocalAddress.put(localWildCardAddress, this);
   }
   
   
/**
 * 
 * @param address
 */
   protected synchronized void registerByRemoteAddress(InetSocketAddress address)
   {
      remoteSocketAddress = address;
      managersByRemoteAddress.put(address, this);
   }
   
   
/**
 * 
 * @param address
 * @throws IOException
 */
   public synchronized void connect(InetSocketAddress address) throws IOException
   {
      if (connected)
      {
         if (socket.getRemoteSocketAddress().equals(address))
            return;
         else
            throw new IOException("socket is already connected");
      }
      
      if (socket == null)
         socket = new Socket();
      
      log.info("connecting to: " + address);
      socket.connect(address);
      registerByRemoteAddress(address);
      connected = true;
      setup();
   }
   
      
/**
 * 
 * @return
 */
   public synchronized void registerServerSocket(ServerSocket serverSocket) throws IOException
   {
      if (this.serverSocket != null && this.serverSocket != serverSocket)
      {
         log.error("[" + id + "]: " + "attempt to register a second server socket");
         log.error("current server socket: " + this.serverSocket.toString());
         log.error("new server socket:     " + serverSocket.toString());
         throw new IOException("attempt to register a second server socket");
      }
      
      log.info("[" + id + "]: " + "server socket registering: " + serverSocket.toString());
      this.serverSocket = serverSocket;
      incrementReferences();
   }
      
      
/**
 * 
 * @param serverSocket
 * @throws IOException
 */
   public synchronized void unRegisterServerSocket(ServerSocket serverSocket) throws IOException
   {
      if (this.serverSocket != serverSocket)
      {
         log.error("server socket attempting unregister but is not registered");
         throw new IOException("server socket is not registered");
      }
      
      log.info("server socket unregistering");
      this.serverSocket = null;
      releaseReference();
   }
      
      
/**
 * 
 * @param socket
 * @return
 * @throws IOException
 */
   public synchronized StreamPair registerSocket(VirtualSocket socket) throws IOException
   {
      SocketId localSocketId = socket.getLocalSocketId();
      VirtualSocket currentSocket = (VirtualSocket) socketMap.put(localSocketId, socket);
      
      if (currentSocket != null)
      {
         String errorMessage = "attempting to register socket on currently used port:"
                              + currentSocket.getLocalVirtualPort();
         log.error(errorMessage);
         throw new IOException(errorMessage);
      }
      
      log.info("registering virtual socket on port: " + localSocketId.getPort());
      connectedRemoteSockets.add(socket.getRemoteSocketId());
      incrementReferences();
      
      MultiplexingInputStream is = getAnInputStream(localSocketId, null);
      MultiplexingOutputStream os = new MultiplexingOutputStream(this, socket, socket.getRemoteSocketId());
      
      return new StreamPair(is, os);
   }
   
 
/**
 * 
 * @param socket
 * @throws IOException
 */
   public synchronized void unRegisterSocket(VirtualSocket socket) throws IOException
   {
      SocketId localSocketId = socket.getLocalSocketId();
      VirtualSocket currentSocket = (VirtualSocket) socketMap.remove(localSocketId);
      
      if (currentSocket == null)
      {
         String errorMessage = "attempting to unregister unrecognized socket";
         log.error(errorMessage);
         throw new IOException(errorMessage);
      }
      
      log.info("unregistering virtual socket on port: " + localSocketId.getPort());
      connectedRemoteSockets.remove(socket.getRemoteSocketId());
      releaseReference();
   }
   
   
 /**
  *  FIXME Comment this
  * 
  *
  */
    public synchronized void registerRemoteServerSocket()
    {
       if (remoteServerSocketRegistered)
          log.error("duplicate remote server socket registration");
       else
       {
          remoteServerSocketRegistered = true;
          incrementReferences();
       }
    }  
  
    
/**
 *  FIXME Comment this
 * 
 *
 */
   public synchronized void unRegisterRemoteServerSocket()
   {
      if (!remoteServerSocketRegistered)
         log.error("no remote server socket is registered");
      else
      {
         remoteServerSocketRegistered = false;
         releaseReference();
      }
   }
   
   
/**
 * 
 * @param socketId
 * @param socket TODO
 * @return
 * @throws IOException
 */
   public MultiplexingInputStream getAnInputStream(SocketId socketId, VirtualSocket socket) throws IOException
   {
      PipedOutputStream pos = new PipedOutputStream();
      MultiplexingInputStream  mis = new MultiplexingInputStream(pos, socket);
      outputStreamMap.put(socketId, pos);
      return mis;
   }

   
/**
 * 
 * @param socketId
 * @return
 * TODO: isn't used
 */
   public boolean isConnectedRemoteSocket(SocketId socketId)
   {
      return connectedRemoteSockets.contains(socketId);
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   public synchronized boolean isServerSocketRegistered()
   {
      return serverSocket != null;  
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   public synchronized ServerSocket getServerSocket()
   {
      return serverSocket;
   }
   
   
/**
 * 
 * @param socketId
 * @return
 */
   public OutputStream getOutputStreamByLocalSocket(SocketId socketId)
   {
      return (OutputStream) outputStreamMap.get(socketId);
   }
   
   
/**
 * 
 * @return
 */
   public Collection getAllOutputStreams()
   {
      return outputStreamMap.values();
   }
   
   
/**
 * 
 * @param socketId
 * @return
 */
   public VirtualSocket getSocketByLocalPort(SocketId socketId)
   {
      return (VirtualSocket) socketMap.get(socketId);
   }
   
   
/**
 * 
 * @return
 */
   public OutputMultiplexor getOutputMultiplexor()
   {
      return outputMultiplexor;
   }
   
   
/**
 * 
 * @return
 */
   public Protocol getProtocol()
   {
      return protocol;
   }
   
  
/**
 * 
 * @return
 */
   public Socket getSocket()
   {
      return socket;
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   public boolean isBound()
   {
      return bound;
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   public boolean isConnected()
   {
      return connected;
   }
   
/**
 * 
 *
 */
   public synchronized void shutdown()
   {
      new ShutdownThread().start();
      
      if (isConnected())
         protocol.notifyOfManagerShutdown();
   }
   
   
/**
 * 
 */
   protected class ShutdownThread extends Thread
   {
      public void run()
      {
         log.info(socket.toString() + ": manager shutting down");      
         running = false;
         
         // When we get here, we don't know how far setting up progressed.
         if (outputThread != null)
         {
            // give output thread a chance to flush pending messages before shutting down socket
            outputThread.shutdown();
            
            try
            {  
               outputThread.join();
               log.info("manager: joined output thread");
            }
            catch (InterruptedException ignored)
            {
               log.debug("interrupted exception waiting for write thread");
            }  
         }
         
         if (socket != null)
         {
            try 
            {
               socket.close();
               log.info("manager: closed socket");
            }
            catch (IOException e)
            {
               log.error("manager: unable to close socket");
            }
         }
         
         if (backChannelThread != null)
         {
            backChannelThread.shutdown();      
            
            try
            {   
               backChannelThread.join();
               log.info("manager: joined back channel thread");
            }
            catch (InterruptedException ignored)
            {
               log.debug("manager: interrupted exception waiting for back channel thread");
            }
         }
         
         if (inputThread != null)
         {
            inputThread.shutdown();
            
            try
            {  
               inputThread.join();
               log.info("manager: joined input thread");
            }
            catch (InterruptedException ignored)
            {
               log.debug("manager: interrupted exception waiting for read thread");
            }    
         }
         
         log.info("manager shut down");
      }
   }

   
     
/**
 * 
 */
   public static class StreamPair
   {
      MultiplexingInputStream is;
      MultiplexingOutputStream os;
      
      public StreamPair(MultiplexingInputStream is, MultiplexingOutputStream os)
      {
         this.is = is;
         this.os = os;
      }
      
      public MultiplexingInputStream getInputStream()
      {
         return is;
      }
      
      public MultiplexingOutputStream getOutputStream()
      {
         return os;
      }
   }
}