/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 * Created on Jul 22, 2005
 */
package org.jboss.remoting.transport.multiplex;


import java.io.IOException;
import java.io.OutputStream;
import java.net.SocketException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.jboss.logging.Logger;

/**
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 * 
 * TODO: size limited payloads
 */
public class OutputMultiplexor
{
   protected static final Logger log = Logger.getLogger(OutputMultiplexor.class);
   protected static final int MARKER = 255;
   
   /** The queue of messages to be processed by the output thread */
//   private LinkedQueue writeQueue;
   private List writeQueue = Collections.synchronizedList(new LinkedList());
   private SocketId previousDestinationId = SocketId.INITIAL_OUTPUT_SOCKET_ID;
   private int markerCount = 0;

   private MultiplexingManager manager;
   private OutputStream os;
   private Message pendingMessage;



/**
 * @param os TODO
 *
 */
   protected OutputMultiplexor(MultiplexingManager manager, OutputStream os)
   {
      this.manager = manager;
      this.os = os;
   }
   
   
   public OutputThread getAnOutputThread()
   {
      return new OutputThread();
   }

   
/**
 * 
 * @param socketId
 * @param content
 * @throws InterruptedException
 */
   public void write(SocketId socketId, byte[] content) throws IOException
   {
      log.debug("OutputMultiplexor.write(): queueing " + content.length + " bytes for socket " + socketId.getPort());
      
      synchronized(writeQueue)
      {
         writeQueue.add(new Message(socketId, content));
         writeQueue.notifyAll();
      }
   }


/** 
*
*/
   class OutputThread extends StoppableThread
   {
      
      private final Logger log = Logger.getLogger(OutputMultiplexor.OutputThread.class);
      
      private int dataOutCount = 0;
      private boolean socketIsOpen = true;
      private boolean trace;
      
      
      public OutputThread()
      {
         trace = log.isTraceEnabled();
      }
      
      
      /**
       * 
       */
      public void shutdown()
      {
         super.shutdown();
         interrupt();
      }
      
      
      protected void doInit()
      {
         log.info("output thread starting");
      }
      
      
      protected void doRun()
      {
         while (socketIsOpen && (!writeQueue.isEmpty() || isRunning()))
         {  
            try
            {
               synchronized(writeQueue)
               {
                  if (writeQueue.isEmpty())
                     writeQueue.wait();
                  
                  pendingMessage = (Message) writeQueue.remove(0);
               }
               
               SocketId remoteSocketId =  pendingMessage.getSocketid();
               int length = pendingMessage.getContent().length; // encode() sets pendingMessage to null
               encode(remoteSocketId, pendingMessage.getContent(), os);
               log.debug("output thread wrote: " + length + " bytes to socket " + remoteSocketId.getPort());
               dataOutCount += length;
               
               if (interrupted()) // outside of writeQueue.take()
                  throw new InterruptedException();
            }
            catch (InterruptedException e)
            {
               handleError("output thread: interrupted", e);
            }
            catch (SocketException e)
            {
               handleError("output thread: socket exception", e);
            }
            catch (IOException e)
            {
               handleError("output thread: i/o error", e);
            }
         }
         
         log.debug("output thread: socketIsConnected: " + socketIsOpen);
         log.debug("output thread: writeQueue.isEmpty(): " + writeQueue.isEmpty());
         log.debug("output thread: running: " + running);
         log.debug("output thread: pendingMessage ==  " + pendingMessage);
      }
      
      
      /**
       * 
       */
      protected void doShutDown()
      {
         //           try
         //           {
         //              os.close();
         //           }
         //           catch (Exception ignored)
         //           {
         //              if (trace)
         //                 log.trace(ignored.getMessage(), ignored);
         //           }
         
         //    TODO: work out
         //           if (terminatedOnError)
         //              multiplexor.shutdown(); 
         
         log.info("output thread: data bytes out: " + dataOutCount);
         log.info("output thread shutting down");
         
         try
         {
            while (pendingMessage != null)
            {
               log.info("waiting for encode() to write final message");
               Thread.sleep(1000);
            }
         }
         catch (InterruptedException ignored)
         {
            log.debug(ignored);
         }
      }
      

   /**
    *
    * @param bytes
    * @param os
    * @throws IOException
    */
      protected void encode(SocketId destinationId, byte[] bytes,  OutputStream os) throws IOException
      {
         try
         {
            if (!destinationId.equals(previousDestinationId))
            {
               if (log.isDebugEnabled())
               {
                  log.debug("encode(): writing: MARKER");
                  byte[] byteArray = destinationId.toByteArray();
                  for (int i = 0; i < byteArray.length; i++)
                     log.debug("encode(): writing: " + byteArray[i]);
                  log.debug("encode(): writing: MARKER");
               }
               
               previousDestinationId = destinationId;
               
               os.write(MARKER);
               os.write(destinationId.toByteArray());
               os.write(MARKER);
            }
            
            int length = bytes.length;
            
            log.debug("encode(): writing " + length + " bytes to socketId " + destinationId.getPort() );
            
            for (int i = 0; i < length; i++)
            {
               int b = 0xff & bytes[i];
               
               if (b != MARKER)
               {
                  if (markerCount > 0) log.debug("wrote " + markerCount + " MARKER's");
                  log.debug("writing byte: " + (int) b);
                  markerCount = 0;
               }
               else
                  markerCount++;
               
               os.write(b);
               
               if (b == MARKER)
                  os.write(MARKER);
            }
            
            //        os.flush();
            
            if (markerCount > 0)
               log.debug("wrote " + markerCount + " MARKER's");
            
            markerCount = 0;
         }
         finally
         {
            pendingMessage = null;
         }
      }
      
      
        
      /**
       * 
       * @param message
       * @param e
       */
      protected void handleError(String message, Throwable e)
      {  
//       if (!socket.isConnected() || socket.isInputShutdown() || socket.isClosed())
         if (e instanceof SocketException)
         {
//            log.debug(message, e);
//            log.info("output thread: socket connected:        " + socket.isConnected());
//            log.info("output thread: socket input shut down:  " + socket.isInputShutdown());
//            log.info("output thread: socket output shut down: " + socket.isOutputShutdown());
//            log.info("output thread: socket closed:           " + socket.isClosed());
            socketIsOpen = false;
            super.shutdown();
            return;
         }
         
         if (log != null)
         {
            //              if (e instanceof InterruptedException || e instanceof IOException)
            //              {
            //                 if (trace)
            //                    log.trace(error, e);
            //              }
            //              else
            log.debug(message, e);
         }
         
         //    TODO
         //           running = false;
         //           terminatedOnError = true;
      }
   }
   
   
   
/**
 * 
 */
   private static class Message
   {
      private SocketId socketId;
      private byte[] content;
      
      public Message(SocketId socketId, byte[] content)
      {
         this.socketId = socketId;
         this.content = content;
      }
      
      public SocketId getSocketid()
      {
         return socketId;
      }
      
      public byte[] getContent()
      {
         return content;
      }
   }
   
}
