/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 * Created on Jul 23, 2005
 */
package org.jboss.remoting.transport.multiplex;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.Socket;

import org.jboss.logging.Logger;

/**
 * 
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class Protocol
{
   protected static final Logger log = Logger.getLogger(Protocol.class);
   
   // message types
   public static final int MP_CONNECT 	 	  		   = 0;
   public static final int MP_CONNECTED 	  		   = 1;
   public static final int MP_INPUT_SHUTDOWN 		   = 2;
   public static final int MP_OUTPUT_SHUTDOWN 		   = 3;
   public static final int MP_DISCONNECT	  		   = 4;
   public static final int MP_SERVER_SOCKET_QUERY 	   = 5;
   public static final int MP_REGISTER_REMOTE_SERVER   = 6;
   public static final int MP_UNREGISTER_REMOTE_SERVER = 7;
   public static final int MP_SHUTDOWN_MANAGER		   = 8;
   public static final int MP_ERROR 	 	  		   = 9;
   public static final int MP_TRUE					   = 10;
   public static final int MP_FALSE					   = 11;
   
   private MultiplexingManager manager;
   
   /** InputStream used by connect() */
   private DataInputStream inputStreamForConnect;
   
   /** OutputStream used by connect() */
   private DataOutputStream outputStreamForConnect;
   
   /** InputStream used by acceptConnect() */
   private DataInputStream inputStreamForAcceptConnect;
   
   /** OutputStream used by answerConnect() */
   private DataOutputStream outputStreamForAnswerConnect;
   
   /** InputStream for receiving asynchronous remote messages */
   private DataInputStream backChannelInputStream;
   
   /** InputStream for sending messages to remote backChannelInputStream */
   private DataOutputStream backChannelOutputStream;
   
   
/**
 * 
 * @param manager
 * @throws IOException
 */
   public Protocol (MultiplexingManager manager) throws IOException
   {
      this.manager = manager;
      Socket socket = manager.getSocket();
      
      inputStreamForConnect = new DataInputStream(manager.getAnInputStream(SocketId.PROTOCOL_SOCKET_ID, null));
      outputStreamForAnswerConnect = new DataOutputStream(new MultiplexingOutputStream(manager, SocketId.PROTOCOL_SOCKET_ID));
      
      if (MultiplexingManager.isOnClient())
      {
         outputStreamForConnect = new DataOutputStream(socket.getOutputStream());
         inputStreamForAcceptConnect = new DataInputStream(manager.getAnInputStream(SocketId.SERVER_SOCKET_ID, null));
       }
      else
      {
         outputStreamForConnect = new DataOutputStream(new MultiplexingOutputStream(manager, SocketId.SERVER_SOCKET_ID));
         inputStreamForAcceptConnect = new DataInputStream(socket.getInputStream());
      }
      
      backChannelOutputStream = new DataOutputStream(new MultiplexingOutputStream(manager, SocketId.BACKCHANNEL_SOCKET_ID));     
      backChannelInputStream = new DataInputStream(manager.getAnInputStream(SocketId.BACKCHANNEL_SOCKET_ID, null));;
   }
        
   
/**
 *  FIXME Comment this
 * 
 * @param socketId
 * @return
 * @throws IOException
 */
   public SocketId connect(SocketId socketId) throws IOException
   {
      outputStreamForConnect.write(MP_CONNECT);
      log.info("Protocol.connect(): wrote: CONNECT (" + MP_CONNECT + ")");
      outputStreamForConnect.writeInt(socketId.getPort());
      log.info("Protocol.connect(): wrote port: " + socketId.getPort());
      int messageType = inputStreamForConnect.read();
      log.info("Protocol.connect(): read message type: " + messageType);
      
      switch (messageType)
      {
	      case MP_CONNECTED:
	         int remotePort = inputStreamForConnect.readInt();
	         log.info("Protocol.connect(): read port: " + remotePort);
	         return new SocketId(remotePort);
	         
	      default:
	         log.error("Protocol.connect(): expecting a CONNECTED message: received: " + messageType);
	         throw new IOException("Protocol.connect(): expecting a CONNECTED message: received: " + messageType);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @param is
 * @return
 * @throws IOException
 */
   public static SocketId acceptConnect(DataInputStream is) throws IOException
   {
      int messageType = is.read();
      log.info("Protocol.acceptConnect(): read message type: " + messageType);
      
      switch (messageType)
      {
	      case MP_CONNECT:
	         int remotePort = is.readInt();
	         log.info("Protocol.acceptConnect(): read port: " + remotePort);
	         return new SocketId(remotePort);
	         
	      default:
	         log.error("Protocol.acceptConnect: expecting a CONNECT message: received: " + messageType);
	         throw new IOException("Protocol.acceptConnect: expecting a CONNECT message: received: " + messageType);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 * @throws IOException
 */
   public SocketId acceptConnect() throws IOException
   {
      log.debug("acceptConnect(): available bytes: " + inputStreamForAcceptConnect.available());
      int messageType = inputStreamForAcceptConnect.read();
      log.info("Protocol.acceptConnect(): read message type: " + messageType);
      
      switch (messageType)
      {
	      case MP_CONNECT:
	         int remotePort = inputStreamForAcceptConnect.readInt();
	         log.info("Protocol.acceptConnect(): read port: " + remotePort);      
	         return new SocketId(remotePort);
	         
	      default:
	         log.error("Protocol.acceptConnect: expecting a CONNECT message: received: " + messageType);
	         throw new IOException("Protocol.acceptConnect: expecting a CONNECT message: received: " + messageType);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @param port
 * @throws IOException
 */
   public void answerConnect(int port) throws IOException
   {
      outputStreamForAnswerConnect.write(MP_CONNECTED);
      log.info("Protocol.answerConnect(): wrote: CONNECTED (" + MP_CONNECTED + ")");
      outputStreamForAnswerConnect.writeInt(port);
      log.info("Protocol.answerConnect(): wrote port: " + port);
   }
   
   
   
/**
 *  FIXME Comment this
 * 
 * @param socketId
 */
   public void notifyInputShutdown(SocketId socketId)
   {
      int port = socketId.getPort();
      
      try
      {
         backChannelOutputStream.write(MP_INPUT_SHUTDOWN);
         backChannelOutputStream.writeInt(port);
         log.info("Protocol.disconnect(): wrote: INPUT_SHUTDOWN (" + MP_INPUT_SHUTDOWN + ") for port: " + port);         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.disconnect(): unable to send MP_INPUT_SHUTDOWN message to port: " + port);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @param socketId
 */
   public void notifyOutputShutdown(SocketId socketId)
   {
      int port = socketId.getPort();
      
      try
      {
         backChannelOutputStream.write(MP_OUTPUT_SHUTDOWN);
         backChannelOutputStream.writeInt(port);
         log.info("Protocol.disconnect(): wrote: OUTPUT_SHUTDOWN (" + MP_OUTPUT_SHUTDOWN + ") for port: " + port);         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.disconnect(): unable to send MP_OUTPUT_SHUTDOWN message to port: " + port);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   public boolean isRemoteServerSocketRegistered() throws IOException
   {
      backChannelOutputStream.write(MP_SERVER_SOCKET_QUERY);
      int b = inputStreamForConnect.read();
      return (b == MP_TRUE) ? true : false;
   }
   
   
   // TODO: work on disconnect
   public void disconnect(SocketId socketId)
   {
      int port = socketId.getPort();
      
      try
      {
         backChannelOutputStream.write(MP_DISCONNECT);
         backChannelOutputStream.writeInt(port);
         log.info("Protocol.disconnect(): wrote: DISCONNECT (" + MP_CONNECTED + ") for port: " + port);         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.disconnect(): unable to send DISCONNECT message to port: " + port);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 *
 */
   public void registerRemoteServerSocket()
   {
      try
      {
         backChannelOutputStream.write(MP_REGISTER_REMOTE_SERVER);
         log.info("Protocol.disconnect(): wrote: REGISTER_REMOTE_SERVER (" + MP_REGISTER_REMOTE_SERVER + ")");         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.notifyOfServerSocketShutdown(): unable to send REGISTER_REMOTE_SERVER");
      }
   }
   

/**
 *  FIXME Comment this
 * 
 *
 */
   public void unregisterRemoteServerSocket()
   {
      try
      {
         backChannelOutputStream.write(MP_UNREGISTER_REMOTE_SERVER);
         log.info("Protocol.disconnect(): wrote: UNREGISTER_REMOTE_SERVER (" + MP_UNREGISTER_REMOTE_SERVER + ")");         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.notifyOfServerSocketShutdown(): unable to send UNREGISTER_REMOTE_SERVER");
      }
   }
   
   
   public void notifyOfManagerShutdown()
   {
      try
      {
         backChannelOutputStream.write(MP_SHUTDOWN_MANAGER);
         log.info("Protocol.disconnect(): wrote: SHUTDOWN_MANAGER (" + MP_UNREGISTER_REMOTE_SERVER + ")");         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.notifyOfServerSocketShutdown(): unable to send SHUTDOWN_MANAGER");
      } 
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   public BackChannelThread getBackChannelThread()
   {
      return new BackChannelThread(manager, backChannelInputStream);
   }
   
   
/**
 * 
 */
   class BackChannelThread extends StoppableThread
   {
      
      private MultiplexingManager manager;
      private DataInputStream is;
      
      
      public BackChannelThread(MultiplexingManager manager, DataInputStream is)
      {
         this.manager = manager;
         this.is = is;
      }

      
/**
 * 
 */
      public void shutdown()
      {
         log.debug("back channel thread: beginning shut down");
         super.shutdown();
         
//         try
//         {
//            is.close();
//         }
//         catch (IOException ignored)
//         {
//            log.debug("back channel thread: i/o exception while shutting down");
//         }
         
         interrupt();
      }
      
      
/**
 * 
 */
      protected void doInit()
      {
         log.info("back channel thread starting");
      }

      
/**
 * 
 */
      protected void doRun()
      {   
         int messageType;
         int port;
         VirtualSocket socket;
         
         try
         {
            messageType = is.read();
            log.info("back channel thread: read message type: " + messageType);
            
            switch (messageType)
            {
               	case MP_SERVER_SOCKET_QUERY:
               	   
               	   boolean answer = manager.isServerSocketRegistered();
               	   outputStreamForAnswerConnect.write(answer ? MP_TRUE : MP_FALSE);
               	   break;
               	   
               	   
               	case MP_INPUT_SHUTDOWN:
               	   
            	   port = is.readInt();
            	   log.info("back channel thread: read INPUT_SHUTDOWN for port: " + port);
            	   socket = manager.getSocketByLocalPort(new SocketId(port)); 
            	   
            	   if (socket == null)
            	   {
            	      log.info("back channel thread (INPUT_SHUTDOWN): unable to retrieve socket at port: " + port);
            	   }
            	   else
            	   {
            	      socket.handleRemoteInputShutDown();
            	   }
            	   
            	   break;
            	   
               	case MP_OUTPUT_SHUTDOWN:
               	   
               	   port = is.readInt();
            	   log.info("back channel thread: read OUTPUT_SHUTDOWN for port: " + port);
            	   socket = manager.getSocketByLocalPort(new SocketId(port)); 
            	   
            	   if (socket == null)
            	   {
            	      log.info("back channel thread (OUTPUT_SHUTDOWN): unable to retrieve socket at port: " + port);
            	   }
            	   else
            	   {
            	      socket.handleRemoteOutputShutDown();
            	   }
            	   
            	   break;
               	   
            	case MP_DISCONNECT:
            	   
            	   port = is.readInt();
            	   log.info("back channel thread: read DISCONNECT for port: " + port);
            	   socket = manager.getSocketByLocalPort(new SocketId(port)); 
            	   
            	   if (socket == null)
            	   {
            	      log.info("back channel thread (DISCONNECT): unable to retrieve socket at port: " + port);
            	   }
            	   else
            	   {
            	      socket.handleRemoteDisconnect();
            	   }
            	   
            	   break;

            	case MP_REGISTER_REMOTE_SERVER:
            	   // remote SinglePortVirtualSocketServer is starting up
              	   log.info("back channel thread: read REGISTER_REMOTE_SERVER");
            	   manager.registerRemoteServerSocket();
            	   break;
            	   
            	case MP_UNREGISTER_REMOTE_SERVER:
            	   // remote SinglePortVirtualSocketServer is shutting down
              	   log.info("back channel thread: read UNREGISTER_REMOTE_SERVER");
            	   manager.unRegisterRemoteServerSocket();
            	   break;
            	   
            	case MP_SHUTDOWN_MANAGER:
            	   // remote MultiplexingManager is shutting down
              	   log.info("back channel thread: read SHUTDOWN_MANAGER");
            	   manager.shutdown();
            	   break;
            	   
            	default:
            	   log.error("unexpected message type in back channel thread: " + messageType);
            }
         }
         catch (InterruptedIOException e)
         {
            if (isRunning())
               log.error("back channel thread: i/o interruption");
            else
               log.debug("back channel thread: i/o interruption");
         }
         catch (IOException e)
         { 
            if (isRunning())
            {
               log.error("back channel thread: i/o error");
               running = false;
            }
            else
               log.debug("back channel thread: i/o error");
         }
      }

      
/**
 * 
 */
      protected void doShutDown()
      {
         log.info("back channel thread shutting down");
      }
      
   }
}
