/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 * Created on Jul 23, 2005
 */
package org.jboss.remoting.transport.multiplex;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.nio.channels.ServerSocketChannel;

import org.jboss.logging.Logger;


/**
 * 
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class SinglePortVirtualServerSocket extends ServerSocket
{
   private static final Logger log = Logger.getLogger(SinglePortVirtualServerSocket.class);
   
   private MultiplexingManager manager;
   private Protocol protocol;
   private Socket actualSocket;
   private boolean bound = false;
   private boolean connected = false;
   private boolean closed = false;
   private boolean startedIndependently = false;
   
   
   
//////////////////////////////////////////////////////////////////////////////////////////////////
///        The following constructors duplicate those of the parent class ServerSocket        '///
//////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @throws java.io.IOException
 */
   public SinglePortVirtualServerSocket() throws IOException
   {
   }
   
      
/**
 * 
 * Create a new SinglePortVirtualServerSocket.
 * 
 * @param port
 * @throws IOException
 */
   public SinglePortVirtualServerSocket(int port) throws IOException
   {
      log.info("entering SinglePortVirtualServerSocket(int port)");
      bind(new InetSocketAddress(port));
      
//      manager = MultiplexingManager.getaManager(port);
//      actualSocket = manager.getSocket();
//      bound = true;
//      log.info("calling registerServerSocket()");
//      manager.registerServerSocket(this);
//      protocol = manager.getProtocol();
//      protocol.registerRemoteServerSocket();
//      connected = manager.isConnected();
//      log.info(toString());
 
   }


/**
 * @param port
 * @param backlog
 * @throws java.io.IOException
 */
//   public SinglePortVirtualServerSocket(int port, int backlog) throws IOException
//   {
//      // TODO
//   }

/**
 * @param port
 * @param backlog
 * @param bindAddr
 * @throws java.io.IOException
 */
//   public SinglePortVirtualServerSocket(int port, int backlog, InetAddress bindAddr) throws IOException
//   {
//      // TODO
//   }
   
   
//////////////////////////////////////////////////////////////////////////////////////////////////
///           The following constructor is particular to SinglePortVirtualServerSocket        '///
//////////////////////////////////////////////////////////////////////////////////////////////////
   
/**
 * 
 * Create a new SinglePortVirtualServerSocket.
 * 
 * @param port
 * @throws IOException
 */
   public SinglePortVirtualServerSocket(InetSocketAddress remoteAddress, InetSocketAddress localAddress)
   throws IOException   
   {
      log.info("entering SinglePortVirtualServerSocket(InetSocketAddress remoteAddress, InetSocketAddress localAddress)");
      bind(localAddress);
       
      // If bind() retrieved a connected MultiplexingManager, we are already connected.
      if (!connected)
      {
         try
         {
            connect(remoteAddress);
         }
         catch (IOException e)
         {
            // During bind(), this ServerSocket was registered with manager.
            if (manager != null && manager.getServerSocket() == this)
               manager.unRegisterServerSocket(this);
            
            log.error(e);
            throw e;
         }
      }
      
//      manager = MultiplexingManager.getaManager(MultiplexingManager.ON_CLIENT, localAddress.getPort());
//      manager.registerServerSocket(this);
//      protocol = manager.getProtocol();
//      protocol.registerRemoteServerSocket();
//      actualSocket = manager.getSocket();
//      bound = true;
//      connected = true;
   }
      
      
//////////////////////////////////////////////////////////////////////////////////////////////////
///                  The following methods are required of any ServerSocket                   '///
//////////////////////////////////////////////////////////////////////////////////////////////////
   
   /*
     ok: public void bind(SocketAddress endpoint) throws IOException;
         public void bind(SocketAddress endpoint, int backlog) throws IOException;
     ok: public InetAddress getInetAddress();
     ok: public int getLocalPort();
     ok: public SocketAddress getLocalSocketAddress();
     ok: public Socket accept() throws IOException;
     ok: public void close() throws IOException;
     ok: public ServerSocketChannel getChannel();
     ok: public boolean isBound();
     ok: public boolean isClosed();
         public void setSoTimeout(int timeout) throws SocketException;
         public int getSoTimeout() throws IOException;
         public void setReuseAddress(boolean on) throws SocketException;
         public boolean getReuseAddress() throws SocketException;
     ok: public String toString();
         public void setReceiveBufferSize(int size) throws SocketException;
         public int getReceiveBufferSize() throws SocketException;
   */
   
/**
 * 
 */
   public Socket accept() throws IOException
   {
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (!isBound())
         throw new SocketException("Socket is not bound yet");
	
      SocketId clientPort = protocol.acceptConnect();
      log.debug("clientPort:  " + clientPort.getPort());
      VirtualSocket virtualSocket = new VirtualSocket(manager, clientPort);
      int localPort = virtualSocket.getLocalVirtualPort();
      protocol.answerConnect(localPort);
      return virtualSocket;
   }
   

/**
 * 
 */
   public void bind(SocketAddress socketAddress) throws IOException
   {
      bind(socketAddress, 0);
   }
   
   
/**
 * 
 */
   public void bind(SocketAddress socketAddress, int backlog) throws IOException
   {
      if (backlog != 0)
         throw new SocketException("bind(): backlog not implemented");
      
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (isBound())
         throw new SocketException("Already bound");
      
      if (socketAddress == null)
         socketAddress = new InetSocketAddress(0);
      
      if (!(socketAddress instanceof InetSocketAddress))
         throw new IllegalArgumentException("Unsupported address type");
      
      InetSocketAddress inetSocketAddress = (InetSocketAddress) socketAddress;
      
      if (inetSocketAddress.isUnresolved())
         throw new SocketException("Unresolved address");
      
      SecurityManager security = System.getSecurityManager();
      
      if (security != null)
         security.checkListen(inetSocketAddress.getPort());
      
      manager = MultiplexingManager.getaManagerByLocalAddress(inetSocketAddress);
      actualSocket = manager.getSocket();
      bound = true;
      manager.registerServerSocket(this);
      
      if (manager.isConnected())
      {
         protocol = manager.getProtocol();
         protocol.registerRemoteServerSocket();
         connected = true;
      }
      
      log.info(toString());
   }
   
   
/**
 * 
 */
   public void close() throws IOException
   {
      if (isClosed())
         return;
      
      closed = true;
      manager.unRegisterServerSocket(this);
      protocol.unregisterRemoteServerSocket();
   }
   
   
/**
 * 
 */
   public InetAddress getInetAddress()
   {
      if (!isBound())
         return null;
      
      return actualSocket.getLocalAddress();
   }
   
   
/**
 * 
 */
   public int getLocalPort()
   {
      if (!isBound())
         return -1;
      
      return actualSocket.getLocalPort();  
   }
   
   
/**
 * 
 */
   public ServerSocketChannel getChannel()
   {
      return null;
   }
   
   
/**
 * 
 */
   public SocketAddress getLocalSocketAddress()
   {
      if (!isBound())
         return null;
      
      return actualSocket.getLocalSocketAddress();  
   }
      
      
/**
 * 
 */
   public boolean isBound()
   {
      return bound;
   }
   
/**
 * 
 */
   public boolean isClosed()
   {
      return closed;
   }
   

/**
 * 
 */
   public String toString()
   {
      StringBuffer answer = new StringBuffer().append("SinglePortVirtualServerSocket[");
      
      if (!isBound())
         return answer.append("unbound]").toString();
      
      answer.append("local address=" + getInetAddress() +  ":" + getLocalPort());
      
      if (!isConnected())
         return answer.append("]").toString();
      
      return answer.append(", remote address=" + actualSocket.getInetAddress() +  
                           ":" + actualSocket.getPort() + "]").toString();
   }
   
   
//////////////////////////////////////////////////////////////////////////////////////////////////
///             The following methods are specific to SinglePortVirtualServerSocket           '///
//////////////////////////////////////////////////////////////////////////////////////////////////

   /**
    * Connects this socket to the server.
    *
    * @param	socketAddress the <code>SocketAddress</code>
    * @throws	IOException if an error occurs during the connection
    * @throws  IllegalArgumentException if endpoint is null or is a
    *          SocketAddress subclass not supported by this socket
    */
   public void connect(SocketAddress socketAddress) throws IOException
   {
	connect(socketAddress, 0);
   }
   
   
/**
 * Connects this socket to the server with a specified timeout value.
 * A timeout of zero is interpreted as an infinite timeout. The connection
 * will then block until established or an error occurs.
 *
 * @param	socketAddress the <code>SocketAddress</code>
 * @param	timeout  the timeout value to be used in milliseconds.
 * @throws	IOException if an error occurs during the connection
 * @throws	SocketTimeoutException if timeout expires before connecting
 * @throws  IllegalArgumentException if endpoint is null or is a
 *          SocketAddress subclass not supported by this socket
 */
   public synchronized void connect(SocketAddress socketAddress, int timeout) throws IOException
   {
      if (timeout > 0)
         throw new SocketException("timeout not yet implemented");
      
      if (socketAddress == null)
         throw new IllegalArgumentException("connect: The address can't be null");
      
      if (timeout < 0)
         throw new IllegalArgumentException("connect: timeout can't be negative");
      
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (!(socketAddress instanceof InetSocketAddress))
         throw new IllegalArgumentException("Unsupported address type");
      
      InetSocketAddress inetSocketAddress = (InetSocketAddress) socketAddress;
      
      // Binding entails getting a MultiplexingManager, which might already be connected,
      // thereby putting the ServerSocket in a connected state.  Therefore, we allow connect() to be
      // called on a connected ServerSocket, as long as the host to which it is to be connected
      // is the same as the host to which it is already connected.
      if (isConnected())
            if (getRemoteAddress().equals(inetSocketAddress.getAddress()))
               return;
            else
               throw new SocketException("already connected");
      
      if (manager == null)
      {
         manager = MultiplexingManager.getaManagerByRemoteAddress(inetSocketAddress);
      }
   
      actualSocket = manager.getSocket();
      protocol = manager.getProtocol();
      
      // If the manager is not connected, we need to connect to the remote MultiPortVirtualServerSocket, which
      // will create a real socket and a MultiplexingManager to wrap it. 
      if (!manager.isConnected())
      {
         manager.connect(inetSocketAddress);
         protocol = manager.getProtocol();
         protocol.connect(SocketId.SERVER_SOCKET_ID);
      }

      protocol.registerRemoteServerSocket();
      
      if (!isBound())
      {
         log.info("calling registerServerSocket()");
         manager.registerServerSocket(this);
         bound = true;
      }
      
      connected = true;
      log.info(toString());
   }  
   
   
/**
 * Returns the connected state of the socket.
 */
   public synchronized boolean isConnected()
   {
      return connected;
   }

   
   public InetAddress getRemoteAddress()
   {
      return actualSocket.getInetAddress();
   }
      
}
