/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 * Created on Jul 22, 2005
 */
package org.jboss.remoting.transport.multiplex;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketImpl;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;

import org.jboss.logging.Logger;

/**
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class VirtualSocket extends Socket
{
   protected static final Logger log = Logger.getLogger(VirtualSocket.class);
   
   private MultiplexingManager manager;
   private Protocol protocol;
   private Socket actualSocket;
   private SocketId remoteSocketId;
   private SocketId localSocketId;
   private MultiplexingInputStream inputStream;
   private MultiplexingOutputStream outputStream;
   
   private boolean connected = false;
   private boolean bound = false;
   private boolean inputShutdown = false;
   private boolean outputShutdown = false;
   private boolean receivedDisconnectMessage = false;
   private boolean closed = false;
   
   
   
   public VirtualSocket(MultiplexingManager manager, SocketId remoteSocketId) throws IOException
   {
      this.manager = manager;
//      manager.incrementReferences();
      this.actualSocket = manager.getSocket();
      this.remoteSocketId = remoteSocketId;
      protocol = manager.getProtocol();
      localSocketId = new SocketId();
      MultiplexingManager.StreamPair streamPair = manager.registerSocket(this);
      inputStream = streamPair.getInputStream();
      outputStream = streamPair.getOutputStream();
      bound = true;
      connected = true;
      log.info("created virtual socket on port: " + localSocketId.getPort());
   }
   
   
//////////////////////////////////////////////////////////////////////////////////////////////////
///           The following constructors duplicate those of the parent class Socket           '///
//////////////////////////////////////////////////////////////////////////////////////////////////
   
/**
 * 
 */
   public VirtualSocket()
   {
      log.info("created unbound virtual socket");
   }
   
   
/**
 * 
 */
   public VirtualSocket(String host, int port) throws UnknownHostException, IOException
   {
      InetSocketAddress address = null;
      
      if (host == null)
         address = new InetSocketAddress(InetAddress.getByName(null), port);
      else
         address = new InetSocketAddress(host, port);
      
      connect(address);
      log.info("created virtual socket on port: " + localSocketId.getPort());
   }
   
   
/**
 * @param host
 * @param port
 * @param stream
 * @throws java.io.IOException
 */
   public VirtualSocket(String host, int port, boolean stream) throws IOException
   {
      if (!stream)
      {
         throw new SocketException("Deprecated: use DataGramSocket instead of stream = false");
      }
      
      InetSocketAddress address = null;
      
      if (host == null)
         address = new InetSocketAddress(InetAddress.getByName(null), port);
      else
         address = new InetSocketAddress(host, port);
      
      connect(address);
      log.info("created virtual socket on port: " + localSocketId.getPort());
   }
   
/**
 * @param address
 * @param port
 * @throws java.io.IOException
 */
   public VirtualSocket(InetAddress address, int port) throws IOException
   {
      connect(new InetSocketAddress(address, port));
      log.info("created virtual socket on port: " + localSocketId.getPort());
   }
   
/**
 * @param host
 * @param port
 * @param stream
 * @throws java.io.IOException
 */
   public VirtualSocket(InetAddress host, int port, boolean stream) throws IOException
   {
      if (!stream)
      {
         throw new SocketException("Deprecated: use DataGramSocket instead of stream = false");
      }

      connect(new InetSocketAddress(host, port));
      log.info("created virtual socket on port: " + localSocketId.getPort());
   }
   
/**
 * @param impl
 * @throws java.net.SocketException
 */
   public VirtualSocket(SocketImpl impl) throws SocketException
   {
      throw new SocketException("VirtualSocket does not use SocketImpl");
   }
   
/**
 * @param host
 * @param port
 * @param localAddr
 * @param localPort
 * @throws java.io.IOException
 */
   public VirtualSocket(String host, int port, InetAddress localAddr, int localPort) throws IOException
   {
      this(InetAddress.getByName(host), port, localAddr, localPort);
   }
   
/**
 * @param address
 * @param port
 * @param localAddr
 * @param localPort
 * @throws java.io.IOException
 */
   public VirtualSocket(InetAddress address, int port, InetAddress localAddr, int localPort) throws IOException
   {
      this();
      bind(new InetSocketAddress(localAddr, localPort));
      connect(new InetSocketAddress(address, port));
      log.info("created virtual socket on port: " + localSocketId.getPort());
   }
   
   
   
//////////////////////////////////////////////////////////////////////////////////////////////////
///                     The following methods are required of all Socket's                    '///
//////////////////////////////////////////////////////////////////////////////////////////////////
   
   
   /*************************************************************************************************
    * public methods in Socket class
    * 
      ok: public void connect(SocketAddress endpoint) throws IOException;
          public void connect(SocketAddress endpoint, int timeout) throws IOException;
      ok: public void bind(SocketAddress bindpoint) throws IOException;
      ok: public InetAddress getInetAddress();
      ok: public InetAddress getLocalAddress();
      ok: public int getPort();
      ok: public int getLocalPort();
      ok: public SocketAddress getRemoteSocketAddress();
      ok: public SocketAddress getLocalSocketAddress();
      ok: public SocketChannel getChannel();
      ok: public InputStream getInputStream() throws IOException;
      ok: public OutputStream getOutputStream() throws IOException;
          public void setTcpNoDelay(boolean on) throws SocketException;
          public boolean getTcpNoDelay() throws SocketException;
          public void setSoLinger(boolean on, int linger) throws SocketException;
          public int getSoLinger() throws SocketException;
          public void sendUrgentData(int data) throws IOException;
          public void setOOBInline(boolean on) throws SocketException;
          public boolean getOOBInline() throws SocketException;
          public void setSoTimeout(int timeout) throws SocketException;
          public int getSoTimeout() throws SocketException;
          public void setSendBufferSize(int size) throws SocketException;
          public int getSendBufferSize() throws SocketException;
          public void setReceiveBufferSize(int size) throws SocketException;
          public int getReceiveBufferSize() throws SocketException;
          public void setKeepAlive(boolean on) throws SocketException;
          public boolean getKeepAlive() throws SocketException;
          public void setTrafficClass(int tc) throws SocketException;
          public int getTrafficClass() throws SocketException;
          public void setReuseAddress(boolean on) throws SocketException;
          public boolean getReuseAddress() throws SocketException;
      ok: public void close() throws IOException;
      ok: public void shutdownInput() throws IOException;
      ok: public void shutdownOutput() throws IOException;
      ok: public String toString();
      ok: public boolean isConnected();
      ok: public boolean isBound();
      ok: public boolean isClosed();
      ok: public boolean isInputShutdown();
      ok: public boolean isOutputShutdown();
   *************************************************************************************************/
   
   
/**
 * Binds the socket to a local address.
 * <P>
 * If the address is <code>null</code>, then the system will pick up
 * an ephemeral port and a valid local address to bind the socket.
 * 
 * TODO: security
 *
 * @param	address the <code>SocketAddress</code> to bind to
 * @throws	IOException if the bind operation fails, or if the socket is already bound.
 * @throws  IllegalArgumentException if bindpoint is a
 *          SocketAddress subclass not supported by this socket
 */
   public synchronized void bind(SocketAddress address) throws IOException
   {
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (isBound())
         throw new SocketException("Already bound");
      
      if (address != null && (!(address instanceof InetSocketAddress)))
         throw new IllegalArgumentException("Unsupported address type");
      
      InetSocketAddress inetAddress = (InetSocketAddress) address;
      
      if (inetAddress != null && inetAddress.isUnresolved())
         throw new SocketException("Unresolved address");
      
//      if (address == null)
//         localSocketId = new SocketId();
//      else
//         //            getImpl().bind(inetAddress.getAddress(),
//         //                  inetAddress.getPort());
//         localSocketId = new SocketId(inetAddress.getPort());
      
      manager = MultiplexingManager.getaManagerByLocalAddress(inetAddress);
      localSocketId = new SocketId();
      log.info("bound virtual socket to port: " + localSocketId.getPort());
      bound = true;
   }

      
/**
 * 
 */
   public synchronized void close() throws IOException
   {
      if (closed)
         return;
      
      closed = true;
      
      if (!receivedDisconnectMessage)
         protocol.disconnect(localSocketId);
      
      inputStream.close();
      inputStream = null;
      
      outputStream.flush();
      outputStream.close();
      outputStream = null;
      
      manager.unRegisterSocket(this);
      log.info("virtual socket closed on port: " + localSocketId.getPort());
   } 
   
   
   /**
    * Connects this socket to the server.
    *
    * @param	endpoint the <code>SocketAddress</code>
    * @throws	IOException if an error occurs during the connection
    * @throws  IllegalArgumentException if endpoint is null or is a
    *          SocketAddress subclass not supported by this socket
    */
   public void connect(SocketAddress endpoint) throws IOException
   {
	connect(endpoint, 0);
   }
   
   
/**
 * Connects this socket to the server with a specified timeout value.
 * A timeout of zero is interpreted as an infinite timeout. The connection
 * will then block until established or an error occurs.
 *
 * @param	socketAddress the <code>SocketAddress</code>
 * @param	timeout  the timeout value to be used in milliseconds.
 * @throws	IOException if an error occurs during the connection
 * @throws	SocketTimeoutException if timeout expires before connecting
 * @throws  IllegalArgumentException if endpoint is null or is a
 *          SocketAddress subclass not supported by this socket
 */
   public synchronized void connect(SocketAddress socketAddress, int timeout) throws IOException
   {
      if (timeout > 0)
         throw new SocketException("timeout not yet implemented");
      
      if (socketAddress == null)
         throw new IllegalArgumentException("connect: The address can't be null");
      
      if (timeout < 0)
         throw new IllegalArgumentException("connect: timeout can't be negative");
      
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (isConnected())
         throw new SocketException("already connected");
      
      if (!(socketAddress instanceof InetSocketAddress))
         throw new IllegalArgumentException("Unsupported address type");
      
      InetSocketAddress inetSocketAddress = (InetSocketAddress) socketAddress;
      
      if (manager == null)
         manager = MultiplexingManager.getaManagerByRemoteAddress(inetSocketAddress);
      
      manager.connect(inetSocketAddress);
      protocol = manager.getProtocol();
      actualSocket = manager.getSocket();
      
      if (!bound)
      {
         localSocketId = new SocketId();
         bound = true;
      }
      
      remoteSocketId = protocol.connect(localSocketId);
      MultiplexingManager.StreamPair streamPair = manager.registerSocket(this);
      inputStream = streamPair.getInputStream();
      outputStream = streamPair.getOutputStream();
      connected = true;
   }
   
   
   /**
 * 
 */
   public synchronized InetAddress getInetAddress()
   {
      if (actualSocket == null)
         return null;
      
      return actualSocket.getInetAddress();
   }
   
   
/**
 * 
 */
   public synchronized InputStream getInputStream() throws IOException
   {
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (isInputShutdown())
         throw new SocketException("Socket input is shutdown");
      
      return inputStream;
   }
   
   
/**
 * Note. Socket.getLocalAddress() returns "wildcard" address for an unbound socket.
 */
   public synchronized InetAddress getLocalAddress()
   {
      if (actualSocket == null)
         return null;
      
      return actualSocket.getLocalAddress();
   }
   
/**
 * 
 */
   public synchronized int getLocalPort()
   {
      if (actualSocket == null)
         return -1;
      
      return actualSocket.getLocalPort();
   }
   
   
/**
 * 
 */
   public SocketAddress getLocalSocketAddress()
   {
      if (actualSocket == null)
         return null;
      
      return actualSocket.getLocalSocketAddress(); 
   }
   
   
/**
 * 
 */
   public synchronized OutputStream getOutputStream() throws IOException
   {
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (isOutputShutdown())
         throw new SocketException("Socket output is shutdown");
      
      // TODO: return distinct output streams? See PlainSocketImpl. 
      //return new SocketOutputStream(this);
      return outputStream;
   }
   
   
/**
 * 
 */
   public synchronized int getPort()
   {
      if (actualSocket == null)
         return 0;
      
      return actualSocket.getPort();
   }
   
   
/**
 * 
 */
   public SocketAddress getRemoteSocketAddress()
   {
      if (actualSocket == null)
         return null;
      
      return actualSocket.getRemoteSocketAddress(); 
   }
   
 /**
 * @return Returns the bound.
 */
   public synchronized boolean isBound()
   {
      return bound;
   }
   
   
/**
 * Returns the closed state of the socket.
 */
   public synchronized boolean isClosed()
   {
      return closed;
   }
   
   
/**
 * Returns the connected state of the socket.
 */
   public synchronized boolean isConnected()
   {
      return connected;
   }

   
/**
 * Returns whether the read-half of the socket connection is closed.
 */
   public synchronized boolean isInputShutdown()
   {
      return inputShutdown;
   }
   
 
/**
 * Returns whether the write-half of the socket connection is closed.
 */
   public synchronized boolean isOutputShutdown()
   {
      return outputShutdown;
   }
      
      
/**
 * 
 */
   public synchronized void shutdownInput() throws IOException
   {
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (!isConnected())
         throw new SocketException("Socket is not connected");
      
      if (isInputShutdown())
         throw new SocketException("Socket input is already shutdown");
      
      inputStream.setEOF();
      inputShutdown = true;
      protocol.notifyInputShutdown(localSocketId);
   }
   
   
/**
* 
*/
   public synchronized void shutdownOutput() throws IOException
   {
      if (isClosed())
         throw new SocketException("Socket is closed");
      
      if (!isConnected())
         throw new SocketException("Socket is not connected");
      
      if (isOutputShutdown())
         throw new SocketException("Socket output is already shutdown");
      
      outputStream.shutdown();
      outputShutdown = true;
      protocol.notifyOutputShutdown(localSocketId);
   }   
   
   
//////////////////////////////////////////////////////////////////////////////////////////////////
///                     The following methods are specific to VirtualSocket's                 '///
//////////////////////////////////////////////////////////////////////////////////////////////////
   public synchronized int getVirtualPort()
   {
      return remoteSocketId.getPort();
   }

   
  
   public synchronized int getLocalVirtualPort()
   {
      return localSocketId.getPort();
   }  
   
   
/**
 * 
 * @return
 */
   public synchronized SocketId getLocalSocketId()
   {
      return localSocketId;
   }
   
   
/**
 * 
 * @return
 */
   public synchronized SocketId getRemoteSocketId()
   {
      return localSocketId;
   }
      
   
/**
 *  FIXME Comment this
 * 
 * @throws IOException
 */
   public void handleRemoteInputShutDown() throws IOException
   {
      // already closed ?
      try
      {
         outputStream.handleRemoteInputShutdown();
      }
      catch (NullPointerException e)
      {
         log.info(e);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @throws IOException
 */
   public void handleRemoteOutputShutDown() throws IOException
   {
      // already closed ?
      try
      {
         inputStream.handleRemoteShutdown();
      }
      catch (NullPointerException e)
      {
         log.info(e);
      }
   }


/**
 * 
 * @throws IOException
 */
   public void handleRemoteDisconnect() throws IOException
   {
//      if (!isConnected() || isClosed())
      
      // if socket becomes connected, go ahead and try to do shutdowns
      if (isClosed())
         return;
      
      log.info("remote virtual socket disconnecting: local port: " + getLocalVirtualPort());
      receivedDisconnectMessage = true;
      
      // already closed ?
      try
      {
         inputStream.handleRemoteShutdown();
      }
      catch (NullPointerException e)
      {
         log.info(e);
      }
      
      // already closed ?
      try
      {
         outputStream.flush();
         outputStream.handleRemoteInputShutdown();
      }
      catch (NullPointerException e)
      {
         log.info(e);
      }
      
      log.info("handleRemoteDisconnect(): done.");
      // TODO
 //     connected = false;
 //     inputShutdown = true;
 //     outputShutdown = true;
   }
   
   
   
//////////////////////////////////////////////////////////////////////////////////////////////////
///                              Synchronized getters and setters                             '///
////////////////////////////////////////////////////////////////////////////////////////////////// 
   
/**
 * @return Returns the actualSocket.
 */
   protected synchronized Socket getActualSocket()
   {
      return actualSocket;
   }
   
/**
 * @param actualSocket The actualSocket to set.
 */
   protected synchronized void setActualSocket(Socket actualSocket)
   {
      this.actualSocket = actualSocket;
   }
   
// in socket API section
///**
// * @return Returns the bound.
// */
//   public synchronized boolean isBound()
//   {
//      return bound;
//   }
   
/**
 * @param bound The bound to set.
 */
   protected synchronized void setBound(boolean bound)
   {
      this.bound = bound;
   }
   
/**
 * @return Returns the manager.
 */
   protected synchronized MultiplexingManager getManager()
   {
      return manager;
   }
   
/**
 * @param manager The manager to set.
 */
   protected synchronized void setManager(MultiplexingManager manager)
   {
      this.manager = manager;
   }
   
   /**
    * @return Returns the protocol.
    */
   protected synchronized Protocol getProtocol()
   {
      return protocol;
   }
   
/**
 * @param protocol The protocol to set.
 */
   protected synchronized void setProtocol(Protocol protocol)
   {
      this.protocol = protocol;
   }
   
/**
 * @param closed The closed to set.
 */
   protected synchronized void setClosed(boolean closed)
   {
      this.closed = closed;
   }
   
/**
 * @param connected The connected to set.
 */
   protected synchronized void setConnected(boolean connected)
   {
      this.connected = connected;
   }
   
/**
 * @param inputShutdown The inputShutdown to set.
 */
   protected synchronized void setInputShutdown(boolean inputShutdown)
   {
      this.inputShutdown = inputShutdown;
   }
   
/**
 * @param inputStream The inputStream to set.
 */
   protected synchronized void setInputStream(MultiplexingInputStream inputStream)
   {
      this.inputStream = inputStream;
   }
   
/**
 * @param localSocketId The localSocketId to set.
 */
   protected synchronized void setLocalSocketId(SocketId localSocketId)
   {
      this.localSocketId = localSocketId;
   }
   
/**
 * @param outputShutdown The outputShutdown to set.
 */
   protected synchronized void setOutputShutdown(boolean outputShutdown)
   {
      this.outputShutdown = outputShutdown;
   }
   
/**
 * @param outputStream The outputStream to set.
 */
   protected synchronized void setOutputStream(MultiplexingOutputStream outputStream)
   {
      this.outputStream = outputStream;
   }
   
/**
 * @param remoteSocketId The remoteSocketId to set.
 */
   protected synchronized void setRemoteSocketId(SocketId remoteSocketId)
   {
      this.remoteSocketId = remoteSocketId;
   }
}
