/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.remoting.transport.servlet.web;

import java.io.IOException;
import javax.management.MBeanServer;
import javax.management.MBeanServerInvocationHandler;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.jboss.logging.Logger;
import org.jboss.mx.util.MBeanServerLocator;
import org.jboss.remoting.transport.servlet.ServletServerInvokerMBean;

/**
 * The servlet that receives the inital http request for the ServletServerInvoker.
 *
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class ServerInvokerServlet extends HttpServlet
{
   private static Logger log = Logger.getLogger(ServerInvokerServlet.class);
   private ServletServerInvokerMBean servletInvoker;

   /**
    * Initializes the servlet.
    */
   public void init(ServletConfig config) throws ServletException
   {
      super.init(config);
      ObjectName localInvokerName = null;
      try
      {
         // See if the servlet is bound to a particular invoker
         String name = config.getInitParameter("invokerName");
         if(name != null)
         {
            localInvokerName = new ObjectName(name);
            log.debug("localInvokerName=" + localInvokerName);
         }
      }
      catch(MalformedObjectNameException e)
      {
         throw new ServletException("Failed to build invokerName", e);
      }

      // Lookup the MBeanServer
      MBeanServer mbeanServer = MBeanServerLocator.locateJBoss();
      if(mbeanServer == null)
      {
         throw new ServletException("Failed to locate the MBeanServer");
      }
      else
      {
         servletInvoker = (ServletServerInvokerMBean)
               MBeanServerInvocationHandler.newProxyInstance(mbeanServer,
                                                             localInvokerName,
                                                             ServletServerInvokerMBean.class,
                                                             false);

      }
   }

   /**
    * Destroys the servlet.
    */
   public void destroy()
   {

   }

   /**
    * Read a MarshalledInvocation and dispatch it to the target JMX object
    * invoke(Invocation) object.
    *
    * @param request  servlet request
    * @param response servlet response
    */
   protected void processRequest(HttpServletRequest request, HttpServletResponse response)
         throws ServletException, IOException
   {
      boolean trace = log.isTraceEnabled();
      if(trace)
      {
         log.trace("processRequest, ContentLength: " + request.getContentLength());
         log.trace("processRequest, ContentType: " + request.getContentType());
      }

      servletInvoker.processRequest(request, response);

   }

   /**
    * Handles the HTTP <code>GET</code> method.
    *
    * @param request  servlet request
    * @param response servlet response
    */
   protected void doGet(HttpServletRequest request, HttpServletResponse response)
         throws ServletException, IOException
   {
      processRequest(request, response);
   }

   /**
    * Handles the HTTP <code>POST</code> method.
    *
    * @param request  servlet request
    * @param response servlet response
    */
   protected void doPost(HttpServletRequest request, HttpServletResponse response)
         throws ServletException, IOException
   {
      processRequest(request, response);
   }

   /**
    * Returns a short description of the servlet.
    */
   public String getServletInfo()
   {
      return "Servlet front to JBossRemoting servlet server invoker.";
   }

}