/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.remoting.transport.socket.ssl;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.util.Map;
import java.util.Properties;
import javax.management.MBeanServer;
import javax.management.MBeanServerInvocationHandler;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.net.ServerSocketFactory;
import javax.net.ssl.SSLServerSocketFactory;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.security.ServerSocketFactoryMBean;
import org.jboss.remoting.transport.socket.SocketServerInvoker;

/**
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class SSLSocketServerInvoker extends SocketServerInvoker implements SSLSocketServerInvokerMBean
{
   protected ServerSocketFactory serverSocketFactory = null;


   public SSLSocketServerInvoker(InvokerLocator locator)
   {
      super(locator);
   }

   public SSLSocketServerInvoker(InvokerLocator locator, Map configuration)
   {
      super(locator, configuration);
   }

   protected ServerSocket createServerSocket(int serverBindPort, int backlog, InetAddress bindAddress) throws IOException
   {
      ServerSocket svrSocket = null;
      Properties props = getProperties();

      /**
       * At this point in time, this can ONLY be set programmatically (via setServerSocketFactory),
       * so if it is set, then will use this.  Otherwise, will try config and then just use new ServerSocket
       * if all else failse.  This is due to JMX based DI (see more below for more).
       */
      if(serverSocketFactory != null)
      {
         svrSocket = serverSocketFactory.createServerSocket(serverBindPort, backlog, bindAddress);
      }
      else
      {

         /**
          * TODO: -TME This is another big hack because of dependancy on JMX within configuration.
          * Have to wait till the mbean server is set before can actually set the server socket
          * factory since it is an mbean (new server's DI will fix all this).  Would prefer this
          * to be in the setup() method...
          * Also, I can't cast the mbean proxy directly to ServerSocketFactory because it is not
          * an interface.  Therefore, have to require that ServerSocketFactoryMBean is used.
          */
         if(props.getProperty("serverSocketFactory") != null)
         {
            try
            {
               String serverSocketFactoryString = props.getProperty("serverSocketFactory");
               if(serverSocketFactoryString != null)
               {
                  MBeanServer server = getMBeanServer();
                  ObjectName serverSocketFactoryObjName = new ObjectName(serverSocketFactoryString);
                  if(server != null)
                  {
                     try
                     {
                        ServerSocketFactoryMBean serverSocketFactory = (ServerSocketFactoryMBean) MBeanServerInvocationHandler.newProxyInstance(server,
                                                                                                                                                serverSocketFactoryObjName,
                                                                                                                                                ServerSocketFactoryMBean.class,
                                                                                                                                                false);
                        svrSocket = serverSocketFactory.createServerSocket(serverBindPort, backlog, bindAddress);
                     }
                     catch(Exception e)
                     {
                        log.error("Error creating mbean proxy for server socket factory for object name: " + serverSocketFactoryObjName, e);
                        throw new IOException("Error createing custom server socket factory.");
                     }
                  }
                  else
                  {
                     log.error("The 'serverSocketFactory' attribute was set with a value, but the MBeanServer reference is null.");
                     throw new IOException("Error creating custom server socket factory.  The invoker does not have a reference to the mbean server.");
                  }
               }
            }
            catch(MalformedObjectNameException e)
            {
               log.error("Error setting the server socket factory due to the attibute value passed not being a valid ObjectName.", e);
               throw new IOException("Error creating custom server socket factory.  The attributed value passed is not a valid object name.");
            }
            catch(NullPointerException e)
            {
               log.error("Error setting the server socket factory due to null pointer exception.", e);
               throw new IOException("Error creating custom server socket factory.");
            }
         }
      }

      if(svrSocket == null)
      {
         log.debug("Creating default server socket.");
         serverSocketFactory = SSLServerSocketFactory.getDefault();
         svrSocket = serverSocketFactory.createServerSocket(serverBindPort, backlog, bindAddress);
      }

      log.debug("Created server socket: " + svrSocket);

      return svrSocket;

   }

   /**
    * Sets the server socket factory for the SocketServerInvoker to use.
    * Can produce normal server socket or ssl server socket, depending
    * on implementation passed.
    *
    * @param serverSocketFactory
    */
   public void setServerSocketFactory(ServerSocketFactory serverSocketFactory)
   {
      this.serverSocketFactory = serverSocketFactory;
   }

}