/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.test.remoting.performance.synchronous;

import java.util.HashMap;
import java.util.Map;
import org.jboss.logging.Logger;
import org.jboss.remoting.Client;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.callback.Callback;
import org.jboss.remoting.callback.HandleCallbackException;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.invocation.NameBasedInvocation;
import org.jboss.remoting.transport.Connector;
import org.jboss.test.remoting.TestUtil;
import org.jboss.test.remoting.performance.PerformanceReporter;

import EDU.oswego.cs.dl.util.concurrent.Latch;
import junit.framework.TestCase;


/**
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 */
public class PerformanceClientTest extends TestCase
{
   private Client client;
   private Connector connector;
   private InvokerLocator locator;

   private Latch serverDoneLock = new Latch();

   // default transport and port
   private String transport = "socket";
   private int port = 9090;

   // performance specific variables
   private int numberOfCalls = 500;
   private int sizeOfPayload = 1024;

   // statics for the specific call methods
   public static final String NUM_OF_CALLS = "numOfCalls";
   public static final String TEST_INVOCATION = "testInvocation";

   private static final Logger log = Logger.getLogger(PerformanceClientTest.class);

   public void init()
   {
      try
      {
         InvokerLocator locator = new InvokerLocator(getTransport() + "://localhost:" + getPort());
         client = new Client(locator, "performance");
         client.connect();
         log.info("Client connected to " + locator.getLocatorURI());
      }
      catch(Exception e)
      {
         log.error(e.getMessage(), e);
      }
   }

   public String getTransport()
   {
      return transport;
   }

   public int getPort()
   {
      return port;
   }

   /**
    * This will be used to create callback server
    *
    * @param port
    * @return
    * @throws Exception
    */
   private InvokerLocator initServer(int port) throws Exception
   {
      if(port < 0)
      {
         port = TestUtil.getRandomPort();
      }
      log.debug("port = " + port);

      connector = new Connector();
      InvokerLocator locator = new InvokerLocator(getTransport() + "://localhost:" + port);
      connector.setInvokerLocator(locator.getLocatorURI());
      connector.create();
      connector.start();
      log.info("Callback server started on " + locator.getLocatorURI());
      return locator;
   }


   public void setUp() throws Exception
   {
      String newTransport = System.getProperty(PerformanceTestCase.REMOTING_TRANSPORT);
      if(newTransport != null && newTransport.length() > 0)
      {
         transport = newTransport;
         log.info("Using transport: " + transport);
      }

      String payloadSize = System.getProperty(PerformanceTestCase.PAYLOAD_SIZE);
      if(payloadSize != null && payloadSize.length() > 0)
      {
         try
         {
            sizeOfPayload = Integer.parseInt(payloadSize);
            log.info("Using payload size: " + sizeOfPayload);
         }
         catch(NumberFormatException e)
         {
            e.printStackTrace();
         }
      }

      String numOfCallsParam = System.getProperty(PerformanceTestCase.NUMBER_OF_CALLS);
      if(numOfCallsParam != null && numOfCallsParam.length() > 0)
      {
         try
         {
            numberOfCalls = Integer.parseInt(numOfCallsParam);
            log.info("Using number of calls: " + numberOfCalls);
         }
         catch(NumberFormatException e)
         {
            e.printStackTrace();
         }
      }

      locator = initServer(-1);
      init();
   }

   public void tearDown() throws Exception
   {
      if(connector != null)
      {
         connector.stop();
         connector.destroy();
         connector = null;
      }
      locator = null;
      if(client != null)
      {
         client.disconnect();
         client = null;
      }
   }

   protected int getNumberOfCalls()
   {
      return numberOfCalls;
   }

   protected int getPayloadSize()
   {
      return sizeOfPayload;
   }

   public void testClientCalls() throws Throwable
   {
      log.debug("running testSynchronousCalls()");
      log.debug("client.getInvoker().getLocator()" + client.getInvoker().getLocator());

      PerformanceCallbackHandler handler = new PerformanceCallbackHandler(client.getSessionId(), serverDoneLock);
      // Need to add callback listener to get callback when the server is done
      client.addListener(handler, locator, client.getSessionId());

      // simple invoke, should return bar
      Object ret = makeInvocation(NUM_OF_CALLS, new Integer(getNumberOfCalls()));

      // create the payload object
      byte[] payload = new byte[getPayloadSize()];
      Payload payloadWrapper = new Payload(payload);

      // THIS IS WHERE TO START THE TIMING
      long startTime = System.currentTimeMillis();

      for(int x = 0; x < getNumberOfCalls(); x++)
      {
         payloadWrapper.setCallNumber(x);
         makeClientCall(TEST_INVOCATION, payloadWrapper);
      }

      long endCallTime = System.currentTimeMillis();
      System.out.println("Time to make all " + getNumberOfCalls() + " is " + (endCallTime - startTime));

      //TODO: -TME Should make this configurable?
      // will make timeout 2 seconds per call
      long timeoutPeriod = 2000 * getNumberOfCalls();
      boolean didComplete = serverDoneLock.attempt(timeoutPeriod);
      long endProcessTime = System.currentTimeMillis();
      if(didComplete)
      {
         int numProcessed = handler.getNumberOfCallsProcessed();
         long totalTime = (endProcessTime - startTime);
         System.out.println("Time for server to process " + numProcessed + " is " + totalTime);
         if(getNumberOfCalls() == numProcessed)
         {
            System.out.println("PASSED - the server processed all calls.");
         }
         else
         {
            System.out.println("FAILED - the server did NOT process all calls.");
         }

         //TODO: -TME - This needs to be replaced by benchmark code reporting
         Map metadata = new HashMap();
         metadata.put("transport", getTransport());
         metadata.put("server total count", String.valueOf(numProcessed));
         metadata.put("number of client calls", String.valueOf(getNumberOfCalls()));
         metadata.put("payload size", String.valueOf(getPayloadSize()));


         PerformanceReporter.writeReport(this.getClass().getName(),
                                         totalTime, getNumberOfCalls(), metadata);

      }
      else
      {
         System.out.println("FAILED - timed out waiting for server to call back when done processing.  Waited for " + timeoutPeriod);
         PerformanceReporter.writeReport("Error in test.  Server never replied that it finished processing.", 0, 0, null);
      }


   }

   protected void makeClientCall(String testInvocation, Payload payloadWrapper) throws Throwable
   {
      makeInvocation(TEST_INVOCATION, payloadWrapper);
   }

   protected Object makeInvocation(String method, Object param) throws Throwable
   {
      Object ret = null;

      if(param != null)
      {
         ret = client.invoke(new NameBasedInvocation(method,
                                                     new Object[]{param},
                                                     new String[]{param.getClass().getName()}),
                             null);
      }
      else
      {
         throw new Exception("To make invocation, must pass a valid, non null, Object as parameter.");
      }

      return ret;
   }

   protected void makeOnewayInvocation(String method, Object param, boolean isClientSide) throws Throwable
   {
      if(param != null)
      {
         client.invokeOneway(new NameBasedInvocation(method,
                                                     new Object[]{param},
                                                     new String[]{param.getClass().getName()}),
                             null,
                             isClientSide);
      }
      else
      {
         throw new Exception("To make oneway invocation, must pass a valid, non null, Object as parameter.");
      }
   }


   public static void main(String[] args)
   {
      PerformanceClientTest test = new PerformanceClientTest();
      try
      {
         test.setUp();
         test.testClientCalls();
         test.tearDown();
      }
      catch(Throwable throwable)
      {
         throwable.printStackTrace();
      }
   }

   public static class PerformanceCallbackHandler implements InvokerCallbackHandler
   {
      private String sessionId;
      private Latch lock;
      private int numberOfCallsProcessed = 0;

      public PerformanceCallbackHandler(String sessionId, Latch lock)
      {
         this.sessionId = sessionId;
         this.lock = lock;
      }

      public int getNumberOfCallsProcessed()
      {
         return numberOfCallsProcessed;
      }

      /**
       * Will take the callback message and send back to client.
       * If client locator is null, will store them till client polls to get them.
       *
       * @param callback
       * @throws org.jboss.remoting.callback.HandleCallbackException
       *
       */
      public void handleCallback(Callback callback) throws HandleCallbackException
      {
            Object ret = callback.getCallbackObject();
            Integer numOfCallsHandled = (Integer) ret;
            System.out.println("Server is done.  Number of calls handled: " + numOfCallsHandled);
            numberOfCallsProcessed = numOfCallsHandled.intValue();
            Object obj = callback.getCallbackHandleObject();
            String handbackObj = (String) obj;
            System.out.println("Handback object should be " + sessionId + " and server called back with " + handbackObj);
            lock.release();
      }
   }
}
