
/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 * Created on Aug 13, 2005
 */
package org.jboss.test.remoting.transport.multiplex;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;

import org.jboss.logging.Logger;
import org.jboss.remoting.transport.multiplex.MultiPortVirtualServerSocket;
import org.jboss.remoting.transport.multiplex.MultiplexingManager;
import org.jboss.remoting.transport.multiplex.VirtualSocket;


/**
 * A MultiplexServer.

 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 * @version $Revision: 1.7 $
 * <p>
 * Copyright (c) 2005
 * </p>
 */

public class MultiplexServer extends PrimeScenarioServer
{
   private static final Logger log = Logger.getLogger(MultiplexServer.class);
   
   private ServerSocket serverSocket;
   private Socket scriptSocket;
   private InputStream scriptInputStream;
   private Socket testSocket;
   
   private ObjectOutputStream os;
   private ObjectInputStream  is;
   
   private byte[] script = new byte[1000];
   private int scriptLength;
   private boolean shutdownServer = false;
   private boolean testOver = false;

   
   protected void runBehaviorTests()
   {      
      log.info("starting behavior tests");
      
      try
      {
         serverSocket = new MultiPortVirtualServerSocket(basicBehaviorServerPort, 50, InetAddress.getByName(basicBehaviorServerHost));
         log.info("MultiplexServer: created MultiPortVirtualServerSocket");
         log.info("localAddress: " + serverSocket.getLocalSocketAddress() + ", local port: " + serverSocket.getLocalPort());
      }
      catch (Exception e)
      {
         log.error("cannot create MultiPortVirtualServerSocket", e);
         e.printStackTrace();
         System.exit(1);
      }
      
      try
      {
         scriptSocket = serverSocket.accept();
         log.info("created script socket");
         scriptInputStream = scriptSocket.getInputStream();
      }
      catch (Exception e)
      {
         log.error("cannot create script socket", e);
         e.printStackTrace();
         System.exit(1);
      }
      
      while (!shutdownServer)
      {
         try
         {
            log.info("MultiplexServer: server socket accepting");
            testSocket = serverSocket.accept();
            log.info("MultiplexServer: created VirtualSocket on port: " + ((VirtualSocket)testSocket).getLocalVirtualPort());
            log.info("localAddress: " + testSocket.getLocalAddress() + ", local port: " + testSocket.getLocalPort());
            log.info("remoteAddress: " + testSocket.getInetAddress() + ", remote port: " + testSocket.getPort());
            
            try
            {
               scriptLength = getScript();
            }
            catch (IOException e)
            {
               log.error("unable to read script");
               System.exit(1);
            }
            
            Thread thread = new BehaviorTestThread(testSocket, script, scriptLength);
            thread.start();
         }
         catch (SocketException ignored)
         {
            // it's because the ServerSocket was closed
            log.info(ignored);
            log.info("shutdownServer: " + shutdownServer);
            log.info("serverSocket:   " + serverSocket.isClosed());
            log.info("testSocket:     " + testSocket.isClosed());
         }
         catch (Exception e)
         {
            log.error("ServerTest.main(): cannot create VirtualSocket", e);
            e.printStackTrace();
            System.exit(1);
         }
      }
      
      log.info("Ending behavior tests");
   }
   
   
   protected int getScript() throws IOException
   {
      try
      {
         scriptLength = scriptInputStream.read();
                  
         for (int i = 0; i < scriptLength; i++)
         {
            script[i] = (byte) scriptInputStream.read();
            
            if (script[i] < 0)
            {
               log.error("end of file reading script");
               throw new IOException("end of file reading script");
            }
         }
      }
      catch (IOException e)
      {
         log.error("error reading script");
         throw e;
      }
      
      return scriptLength;
   }

   
   
   class BehaviorTestThread extends Thread
   {
      private Socket socket;
      private byte[] script;
      private int scriptLength;
      private InputStream  is;
      private OutputStream os;
      private VirtualSocket socketToClient;
      private InputStream is_client;
      private OutputStream os_client;

      
      public BehaviorTestThread(Socket testSocket, byte[] script, int scriptLength)
      {
         this.socket = testSocket;
         this.script = script;
         this.scriptLength = scriptLength;
         
         try
         {
            is = testSocket.getInputStream();
            os = testSocket.getOutputStream();
         }
         catch (IOException e)
         {
            log.error("i/o error creating InputStream or OutputStream", e);
            e.printStackTrace();
         }
      }
      

      public void run()
      {
         log.info("starting behavior test thread");
         log.info("script length: " + scriptLength);
         int b = 0;
         
         try
         {
            for (int i = 0; i < scriptLength; i++)
            {
               log.info("script command: " + script[i]);
               
               switch (script[i])
               {
                  case READ:
                     b = is.read();
                     log.debug("read: " + b);
                     break;
                     
                  case WRITE:
                     os.write(b);
                     break;
                     
                  case SHUTDOWN_INPUT:
                     socket.shutdownInput();
                     break;
                     
                  case SHUTDOWN_OUTPUT:
                     socket.shutdownOutput();
                     break;
                     
                  case SLEEP:
                     Thread.sleep(3000);
                     break;
                     
                  case CLOSE_TEST_SOCKET:
                     socket.close();
                     break;
                     
                  case CLOSE_SCRIPT_SOCKET:
                     scriptSocket.close();
                     break;
                     
                  case CONNECT_TO_CLIENT:
                     
                     try
                     {
                        socketToClient = new VirtualSocket(clientServerSocketHost, socket.getPort());
                        is_client = socketToClient.getInputStream();
                        os_client = socketToClient.getOutputStream();
                     }
                     catch (IOException e)
                     {
                        log.error("unable to create VirtualSocket back to client", e);
                        socketToClient = null;
                     }
                     
                     break;
                     
                  case CONNECT_TO_CLIENT_SINGLE_PORT:
                     
                     try
                     {
                        socketToClient = new VirtualSocket(clientServerSocketHost, clientServerSocketPort);
                        is_client = socketToClient.getInputStream();
                        os_client = socketToClient.getOutputStream();
                     }
                     catch (IOException e)
                     {
                        log.error("unable to create VirtualSocket back to client", e);
                        socketToClient = null;
                     }
                     
                     break;
                     
                     
                  case CONNECT_TO_CLIENT_MULTI_PORT:
                     
                     try
                     {
                        MultiplexingManager.setOnClient(true);
                        socketToClient = new VirtualSocket(clientServerSocketHost, clientServerSocketPort);
                        MultiplexingManager.setOnClient(false);
                        is_client = socketToClient.getInputStream();
                        os_client = socketToClient.getOutputStream();
                     }
                     catch (IOException e)
                     {
                        log.error("unable to create VirtualSocket back to client", e);
                        socketToClient = null;
                     }
                     
                     break;
                     
                  case READ_FROM_CLIENT:
                     
                     if (socketToClient == null)
                        log.error("unable to read from client: socket has not been opened");
                     else
                        b = is_client.read();
                     
                     break;
                     
                  case WRITE_TO_CLIENT:
                     
                     if (socketToClient == null)
                        log.error("unable to write to client: socket has not been opened");
                     else
                        os_client.write(i);
                     
                     break;
                     
                  case CLOSE_CLIENT_SOCKET:
                     
                     if (socketToClient == null)
                        log.error("unable to close socket to client: not open");
                     else
                        socketToClient.close();
                     
                     break;
                     
                  case END_TESTS:
                     socket.close();
                     shutdownServer = true;
                     serverSocket.close();
                     break;
                     
                  default:
                     log.error("unrecognized command: " + script[i]); 
               }
            }
         }
         catch (Exception e)
         {
            log.error("error in behavior test thread", e);
         }
         
         log.info("behavior test thread shutting down");
      }
   }
   
   public static void main(String[] args)
   {
      MultiplexingManager.setOnClient(false);
      MultiplexServer multiplexServer = new MultiplexServer();
      multiplexServer.runPrimeScenario();
      multiplexServer.runBehaviorTests();
   }

}

